/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_palmDb.c,v 1.19 2005/11/17 17:19:21 guich Exp $ */

/* These routines implement the PDB file format on a standard file system.
   The PDB is handled, in SuperWaba, by the waba.io.Catalog class.

   This file contains the platform independent routines.
   The files _palm/_ce/_posix contains platform dependent code.

	Catalog Premisses:

   1. Read only catalogs can have more than one locked records (used when
      loading classes, loading fonts, etc - only by the VM)

   2. Read-write catalogs can only have one record opened at a time (as the
      Catalog.java class works)

   Notes:

   . Locked records have its contents copied into a temp location and are
     directly manipulated

   . When a record is closed, if modified, it is written directly to "disk",
     expanding of shrinking the records as necessary.
*/

#include "superwaba.h"
#include "SW_palmDbPrivate.h"

#define DB_NAME_LENGTH  32       // 31 chars + 1 null terminator
#define DB_IS_OPEN (db != NULL && ((void*)db->fh) != NULL)
#define GET_OFFSET(idx) (78 + ((int32)idx << 3))


// Attributes of a Database
//
#define  dmHdrAttrResDB             0x0001   // Resource database
#define  dmHdrAttrReadOnly          0x0002   // Read Only database
#define  dmHdrAttrAppInfoDirty      0x0004   // Set if Application Info block is dirty
                                             // Optionally supported by an App's conduit
#define  dmHdrAttrBackup            0x0008   // Set if database should be backed up to PC if
                                             // no app-specific synchronization conduit has
                                             // been supplied.
#define  dmHdrAttrOKToInstallNewer  0x0010   // This tells the backup conduit that it's OK
                                             //  for it to install a newer version of this database
                                             //  with a different name if the current database is
                                             //  open. This mechanism is used to update the
                                             //  Graffiti Shortcuts database, for example.
#define  dmHdrAttrResetAfterInstall 0x0020   // Device requires a reset after this database is
                                             // installed.
#define  dmHdrAttrCopyPrevention    0x0040   // This database should not be copied to

#define  dmHdrAttrStream            0x0080   // This database is used for file stream implementation.
#define  dmHdrAttrHidden            0x0100   // This database should generally be hidden from view
                                             //  used to hide some apps from the main view of the
                                             //  launcher for example.
                                             // For data (non-resource) databases, this hides the record
                                             //  count within the launcher info screen.
#define  dmHdrAttrLaunchableData    0x0200   // This data database (not applicable for executables)
                                             //  can be "launched" by passing it's name to it's owner
                                             //  app ('appl' database with same creator) using
                                             //  the sysAppLaunchCmdOpenNamedDB action code.

#define  dmHdrAttrRecyclable        0x0400   // This database (resource or record) is recyclable:
                                             //  it will be deleted Real Soon Now, generally the next
                                             //  time the database is closed.

#define  dmHdrAttrBundle            0x0800   // This database (resource or record) is associated with
                                             // the application with the same creator. It will be beamed
                                             // and copied along with the application.

#define  dmHdrAttrOpen              0x8000   // Database not closed properly


// Record Attributes
//
#define  dmRecAttrDelete         0x80  // delete this record next sync
#define  dmRecAttrDirty          0x40  // archive this record next sync
#define  dmRecAttrBusy           0x20  // record currently in use
#define  dmRecAttrSecret         0x10  // "secret" record - password protected

// Record Header
//
typedef struct
{
   uint32 offset;
   byte attr;
   byte uniqueId[3];
} RecordHeader;

// DataBase Header
//
typedef struct
{
   char name[DB_NAME_LENGTH];
   /** The attribute flags for the database. */
   uint16 attributes;
   /** The application-specific version of the database layout. */
   uint16 version;
   /** The creation date of the database, specified as the number of seconds since
   * 12:00 A.M. on January 1, 1904.
   */
   uint32 creationDate;
   /** The date of the most recent modification of the database, specified as the
   * number of seconds since 12:00 A.M. on January 1, 1904.
   */
   uint32 modificationDate;
   /** The date of the most recent backup of the database, specified as the number
   * of seconds since 12:00 A.M. on January 1, 1904.
   */
   uint32 lastBackupDate;
   /** The modification number of the database. */
   uint32 modificationNumber;
   /** The local offset from the beginning of the database header data to the start
   * of the optional, application-specific appInfo block. This value is set to NULL
   * for databases that do not include an appInfo block.
   */
   uint32 appInfoOffset;
   /** The local offset from the beginning of the PDB header data to the start of the
   * optional, application-specific sortInfo block. This value is set to NULL for
   * databases that do not include an sortInfo block.
   */
   uint32 sortInfoOffset;
   uint32 type; // guich@120
   uint32 creator; // guich@120
   /** Used internally by the Palm OS to generate unique identifiers for records on the
   * Palm device when the database is loaded into the device. For PRC databases, this
   * value is normally not used and is set to 0. For PQA databases, this value is not
   * used, and is set to 0.
   */
   uint32 uniqueIDSeed;
   /** The local chunk ID of the next record list in this database. This is 0 if there
   * is no next record list, which is almost always the case. <b>Important!</b> In
   * SuperWaba, this type of database is not supported!
   */
   uint32 nextRecordListID;
   /** Number of records in this db */
   uint16 numRecords;
} DatabaseHeader;

// builds a linked list of locked records
typedef struct tagRecordList
{
   byte* recPtr;
   struct tagRecordList *next;
   uint16 index;
   uint16 size;
   byte changed;
} RecordList;

// Catalog
//
typedef struct t_Catalog
{
   DatabaseHeader dbh;
   _HANDLE fh;
   uint32 fileSize;
   TCHAR fullPath[MAX_PATH-31];
   int32 mode;
   // current active record
   byte *recordBuf; // buffer used to store the current record
   uint32 sizeofRecordBuf;
   uint16 recordSize;
   uint16 originalRecordSize;
   int32 recordIndex;
   byte recordChanged;
   int32 lockedRecords; // count

   uint32 originalModificationNumber;
   struct tagRecordList *recPtrList[16]; // a mini hashtable to store the gotRecords
   byte isNewRecord;
} Catalog, *CatalogPtr;

byte byteBuf[BYTE_BUF_LEN];
static RecordHeader rhs[16]; // to optimize read/writes
static int32 lastPos = -1, lastPos2 = -1;
static MemHandle getRecord(DmOpenRef dbref, uint16 index, int32 *size, byte queryOnly, int32 ofs, byte* buf);

/*----------------------------------------------------------------storeRecPtr-+
|                                                                             |
+----------------------------------------------------------------------------*/
static byte storeRecPtr(DmOpenRef dbref, byte* mh, uint16 index, uint16 size)
{
   CatalogPtr db = (CatalogPtr)dbref;
   int pos = index & 0xF;
   RecordList *rl = (RecordList *)vmGlobals->xmalloc(sizeof(RecordList));
   if (!rl) return 0;
   xmemzero(rl, sizeof(RecordList));

   rl->recPtr = mh;
   rl->index = index;
   rl->size = size;
   rl->next = db->recPtrList[pos];
   db->recPtrList[pos] = rl;
   return 1;
}

/*------------------------------------------------------------retrieveRecList-+
|                                                                             |
+----------------------------------------------------------------------------*/
static RecordList* retrieveRecList(DmOpenRef dbref, uint16 index)
{
   CatalogPtr db = (CatalogPtr)dbref;
   int pos = index & 0xF;
   RecordList *rl = (RecordList *)db->recPtrList[pos];
   while (rl != NULL)
      if (rl->index == index)
         return rl;
      else
         rl = rl->next;
   return NULL;
}

/*--------------------------------------------------------------destroyRecPtr-+
|                                                                             |
+----------------------------------------------------------------------------*/
static byte destroyRecPtr(DmOpenRef dbref, uint16 index)
{
   CatalogPtr db = (CatalogPtr)dbref;
   int pos = index & 0xF;
   RecordList *rl = (RecordList *)db->recPtrList[pos];
   RecordList *last = NULL;
   while (rl != NULL)
      if (rl->index == index)
      {
         if (last == NULL)
            db->recPtrList[pos] = rl->next; // first item?
         else
            last->next = rl->next;
         xfree(rl->recPtr);
         xfree(rl);
         return 1;
      }
      else
      {
         last = rl;
         rl = rl->next;
      }
   return 0;
}

/*-------------------------------------------------------------destroyRecPtrs-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void destroyRecPtrs(DmOpenRef dbref)
{
   CatalogPtr db = (CatalogPtr)dbref;
   int i;
   for (i = 0; i < 16; i++)
   {
      RecordList *rl = (RecordList *)db->recPtrList[i];
      while (rl != NULL)
      {
         RecordList *next = rl->next;
         xfree(rl->recPtr);
         xfree(rl);
         rl = next;
      }
   }
}

/*------------------------------------------------------------replaceDatapath-+
|                                                                             |
+----------------------------------------------------------------------------*/

// guich@550: lets test this.
static WClassField *dataPathField;
#if defined(PALMOS) && !defined(__GCC32__)
#define lstrlen xstrlen
#define lstrcmp(x,y) xstrncmp(x,y,xstrlen(x))
#endif

static TCHAR* replacePath(TCHAR *buf, char16 *newPath, int32 len, TCHAR *oldPath)
{
   // copy the current dataPath to the buffer, taking care of the slashes
   TCHAR *dst,*src;
   int32 pos;
   dst = buf;
   for (; len--; dst++,newPath++)
      *dst = (*newPath != '/') ? (TCHAR)*newPath : '\\';
   if (*(dst-1) != '\\') *dst++ = '\\'; // add the last slash
   // finds the position for the \ or / in the path
   len = lstrlen(oldPath);
   pos = len-1;
   src = &oldPath[pos];
   for (;pos >= 0 && *src != '/' && *src != '\\'; pos--,src--) ;
   // copies the string
   len -= pos+1;
   for (src++; len-- > 0; src++,dst++)
      *dst = *src;
   // zero-end the string
   *dst = 0;
   return buf;
}

// reads the waba.sys.Settings.dataPath string and, if it's set,
// replace the path from the given filename with the dataPath path.
TCHAR *replaceDataPath(TCHAR *fullPath)
{
   WObject obj,charArray;
   static TCHAR buf[MAX_PATH];

   if (vmGlobals->prefClass == NULL) // vm still not initialized?
      return fullPath;
   if (dataPathField == NULL)
   {
      UtfString fieldName = vmGlobals->createUtfString("dataPath");
      UtfString fieldDesc = vmGlobals->createUtfString("Ljava/lang/String;");
      dataPathField = vmGlobals->getField(vmGlobals->prefClass, &fieldName, &fieldDesc);
      if (dataPathField == NULL)
         return fullPath;
   }
   obj = dataPathField->var.staticVar.obj;
   if (obj == NULL || (charArray=WOBJ_StringCharArrayObj(obj)) == NULL) // if string null, delete it - guich@240_3: first condition added.
      return fullPath;
   else
      return replacePath(buf, (char16 *)WOBJ_arrayStart(charArray), WOBJ_arrayLen(charArray), fullPath);
}
//#endif

/*------------------------------------------------------------------shiftFile-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void shiftFile(DmOpenRef dbref, int32 start, int32 end, int32 len)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (len > 0) // insert bytes at current position?
   {
      while ((end-BYTE_BUF_LEN) > start)
      {
         end -= BYTE_BUF_LEN;
         _myReadAt(db->fh, byteBuf, BYTE_BUF_LEN, end);
         _myWriteAt(db->fh, byteBuf, BYTE_BUF_LEN, end+len);
      }
      _myReadAt(db->fh, byteBuf, end-start, start);
      _myWriteAt(db->fh, byteBuf, end-start, start+len);
   }
   else // delete bytes at current position
   {
      while ((start+BYTE_BUF_LEN) < end)
      {
         _myReadAt(db->fh, byteBuf, BYTE_BUF_LEN, start);
         _myWriteAt(db->fh, byteBuf, BYTE_BUF_LEN, start+len);
         start += BYTE_BUF_LEN;
      }
      _myReadAt(db->fh, byteBuf, end-start, start);
      _myWriteAt(db->fh, byteBuf, end-start, start+len);
   }
}

/*---------------------------------------------------------swapDatabaseHeader-+
|                                                                             |
+----------------------------------------------------------------------------*/
//#if SDL_BYTEORDER == SDL_LIL_ENDIAN
/*
| the previous #IF could be suppressed, b/c, for big endians,
| swapIntXx is #define as SDL_SwapBE (Swap to Big Endians)
*/
static void swapDatabaseHeader(DatabaseHeader *src, DatabaseHeader *dst)
{
   _LOG_DEBUG((LOG_CAT_NLIB, "swapDatabaseHeader(%s)", src->name));
   xmemmove(dst->name, src->name, DB_NAME_LENGTH);   
   dst->attributes = READ_BE_UINT16(&src->attributes);
   dst->version = READ_BE_UINT16(&src->version);
   dst->creationDate = READ_BE_UINT32(&src->creationDate);
   dst->modificationDate = READ_BE_UINT32(&src->modificationDate);
   dst->lastBackupDate = READ_BE_UINT32(&src->lastBackupDate);
   dst->modificationNumber = READ_BE_UINT32(&src->modificationNumber);
   dst->appInfoOffset = READ_BE_UINT32(&src->appInfoOffset);
   dst->sortInfoOffset = READ_BE_UINT32(&src->sortInfoOffset);
   dst->type = READ_BE_UINT32(&src->type);
   dst->creator = READ_BE_UINT32(&src->creator);
   dst->uniqueIDSeed = READ_BE_UINT32(&src->uniqueIDSeed);
   dst->nextRecordListID = READ_BE_UINT32(&src->nextRecordListID);
   dst->numRecords = READ_BE_UINT16(&src->numRecords);
   _LOG_DEBUG((LOG_CAT_NLIB, "swapDatabaseHeader done"));
}
//#else
//#define swapDatabaseHeader(dh) (dh)
//#endif

/*--------------------------------------------------------------getDBFileName-+
|                                                                             |
+----------------------------------------------------------------------------*/
static TCHAR * getDBFileName(DmOpenRef dbref, TCHAR *buf)
{
   CatalogPtr db = (CatalogPtr)dbref;
   TCHAR *cOut = buf;
   TCHAR *cIn;
   TCHAR *dp;
   char *c;
   // add the full path first
   cIn = db->fullPath;
   while (*cIn)
      *cOut++ = *cIn++;
   // add the pdb name
   c = db->dbh.name;
   while (*c)
      *cOut++ = (TCHAR) *c++;
   // add the extension
   c = ".pdb";
   while (*c)
      *cOut++ = (TCHAR) *c++;
   *cOut = 0;
   dp = replaceDataPath(buf); // guich@552_34: use the datapath if provided
   if (dp != buf)
#ifdef UNICODE
      lstrcpy(buf, dp); 
#else
      xstrcpy(buf, dp); 
#endif      
   return buf;
}

/*----------------------------------------------------------------mySplitPath-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void mySplitPath(TCHAR *fullPath, TCHAR *path, TCHAR *name)
{
   TCHAR *c = fullPath;
   TCHAR *bs = NULL;
   while (*c)
   {
      if (*c == DIRSEP) bs = c;
      ++c; // go to end of fullPath
   }
   if (!bs)  // no slash? reading from the root?
   {
      if (path)
      {
         *path++ = DIRSEP;
         *path = 0;
      }
      if (name)
      {
         xmemmove(name, fullPath, (c-fullPath)*sizeof(TCHAR));
         name[c-fullPath] = 0;
      }
   }
   else
   {
      ++bs;
      if (path)
      {
         xmemmove(path, fullPath, (bs-fullPath)*sizeof(TCHAR));
         path[bs-fullPath] = 0;
      }
      if (name)
      {
         xmemmove(name, bs, (c-bs)*sizeof(TCHAR));
         name[c-bs] = 0;
      }
   }
}

/*----------------------------------------------------------------loadCatalog-+
|                                                                             |
+----------------------------------------------------------------------------*/
static CatalogPtr loadCatalog(TCHAR * fullPath, byte createIt, uint32 creator, uint32 type, byte readOnly)
{
   _HANDLE h;
   CatalogPtr db;
   int newFile;
   uint16 att;
   static TCHAR name[DB_NAME_LENGTH];

   h = _myCreateFile(fullPath, createIt, readOnly);
   if (!h) return NULL;


   db = (CatalogPtr)(*vmGlobals->xmalloc)(sizeof(Catalog));
   if (!db) return NULL;
   xmemzero(db,sizeof(Catalog));
   db->dbh.version = 1;
   db->fh = h;
   db->fileSize = _myGetFileSize(db->fh);
   db->recordIndex = -1;

   // xmemzero(name, sizeof(name));
   mySplitPath(fullPath, db->fullPath, name);
   newFile = db->fileSize < 78;
   if (newFile)
   {
      DatabaseHeader swappedDh;
      int16 pad = 0;
      TCHAR *tc;
      char *c;
      if (!createIt)
      {
         _myCloseFile(db->fh);
         xfree(db);
         return NULL;
      }
      // set the fields for the newly created database
      db->dbh.modificationDate = db->dbh.creationDate = _myGetNow();
      db->dbh.lastBackupDate   = db->dbh.creationDate - 365 * 24 * 60 * 60; // creation date minus one year: make sure we will be hotsync'ed!
      db->dbh.creator = creator;
      db->dbh.type = type;
      // set the name
      c = db->dbh.name;
      tc = name;
      while (*tc && *tc != '.')
         *c++ = (char) *tc++;
      *c = 0;
      // write header info into the newly created catalog
      swapDatabaseHeader(&db->dbh,&swappedDh);
      _myWrite(db->fh, &swappedDh, 78);
      _myWrite(db->fh, &pad, 2); // this goes after the table of record headers
      //flush(db->fh);
   }
   else
   {
      _myRead(db->fh, &db->dbh, 78);
      swapDatabaseHeader(&db->dbh,&db->dbh);
      // check if really right file
      if (db->dbh.creator != creator || db->dbh.type != type)
      {
         _myCloseFile(db->fh);
         xfree(db);
         return NULL;
      }
      // was'nt the db properly closed?
      if (db->dbh.attributes & dmHdrAttrOpen)
      {
         // read all record headers and check if they have were left busy
         uint16 count = db->dbh.numRecords;
         int32 offset = 78;
         byte mustSave;
         int i;

         while (count > 16)
         {
            mustSave = 0;
            _myReadAt(db->fh, rhs, 8*16, offset);
            for (i = 0; i < 16; i++)
               if (rhs[i].attr & dmRecAttrBusy) // no need to swapRecordHeader here
               {
                  rhs[i].attr &= ~dmRecAttrBusy;
                  mustSave = 1;
               }
            if (mustSave)
               _myWriteAt(db->fh, rhs, 8*16, offset);
            count -= 16;
            offset += 8*16;
         }
         if (count > 0)
         {
            mustSave = 0;
            _myReadAt(db->fh, rhs, 8*count, offset);
            for (i = 0; i < count; i++)
               if (rhs[i].attr & dmRecAttrBusy)
               {
                  rhs[i].attr &= ~dmRecAttrBusy;
                  mustSave = 1;
               }
            if (mustSave)
               _myWriteAt(db->fh, rhs, 8*count, offset);
         }
      }
      // set the opened attribute
      db->dbh.attributes |= dmHdrAttrOpen;
      if (!readOnly)
      {
         att = READ_BE_UINT16(&db->dbh.attributes);
         _myWriteAt(db->fh, &att, 2, 32);
      }
   }
   db->originalModificationNumber = db->dbh.modificationNumber;
   return db;
}

static int32 ascii2unicode(TCHAR *dst, char *src, int32 max) // palm only
{
#if defined(__GCC32__) || defined (WIN32) || defined (LINUX)
   return vmGlobals->ascii2unicode(dst, src, max);
#else
   int len = xstrlen(src);
   if (max <= len) len = max-1;
   xstrncpy(dst, src, len);
   dst[len] = 0;
   return len;
//   int l1 = xstrlen(src);
//   int l2 = xstrlen(dst);
//   if ((l1+l2) > max) l1 = max-l2;
//   xstrncpy(dst,src,l1);
//   dst[l1] = 0;
//   return l1;
#endif
}

/*-------------------------------------------------------------createFileName-+
|                                                                             |
+----------------------------------------------------------------------------*/
void createFileName(char *appPath, char *file, TCHAR *path, char *ext)
{
   int i;
   i = ascii2unicode(path, appPath, MAX_PATH);
   i += ascii2unicode(&path[i], file, MAX_PATH-i);
   ascii2unicode(&path[i], ext, MAX_PATH-i);
}

/*------------------------------------------------------------createFileName2-+
|                                                                             |
+----------------------------------------------------------------------------*/
#ifdef WINCE // should be nice to test in other platforms where sizeof(TCHAR) == 2
static void createFileName2(TCHAR *appPath, char *file, TCHAR *path, char *ext)
{
   int i;
   i = lstrlen(appPath);
   xmemmove(path, appPath, i*2);
   //i = ascii2unicode(path, appPath, MAX_PATH);
   i += ascii2unicode(&path[i], file, MAX_PATH-i);
   ascii2unicode(&path[i], ext, MAX_PATH-i);
}
#else
#define createFileName2 createFileName
#endif

static TCHAR tempPath[MAX_PATH]; // cannot be hold between other routine calls

/*---------------------------------------------------------myDmCreateDatabase-+
|                                                                             |
+----------------------------------------------------------------------------*/
Err myDmCreateDatabase(char *name, uint32 creator, uint32 type)
{
   CatalogPtr db;
#ifndef PALMOS
   if (name[0] == '\\' || name[0] == '.') // guich@400_57: allow the creation of pdbs in another directory
      createFileName("", name, tempPath, ".pdb");
   else
#endif
      createFileName(getAppPath(), name, tempPath, ".pdb");

   db = loadCatalog(tempPath, 1, creator, type,0);
   if (db == NULL) // could not create catalog?
      return PALM_ERROR;

   myDmCloseDatabase(db); // loadCatalog remained opened
   return PALM_SUCCESS;
}

/////////////////////////////////////////////////////////////////////////////
DmOpenRef myDmOpenDatabase(TCHAR *fileName, uint32 creator, uint32 type)
{
   return loadCatalog(fileName, 0, creator, type,0);
}

/////////////////////////////////////////////////////////////////////////////
DmOpenRef myDmOpenDatabaseReadOnly(TCHAR *fileName, uint32 creator, uint32 type) // guich@310_17: let more than one instance of SW
{
   return loadCatalog(fileName, 0, creator, type,1);
}
/////////////////////////////////////////////////////////////////////////////

typedef struct
{
   TCHAR    *fileName;
   int    fileNameLen;
   uint32    creator;
   uint32    type;
   DmOpenRef foundCatalog;
} FileMatches_Vars;

static byte fileMatches(TCHAR *fileName, void *userVars)
{
   FileMatches_Vars *vars = (FileMatches_Vars *)userVars;
   int len = lstrlen(fileName);
   if (len > vars->fileNameLen && lstrcmp(fileName-vars->fileNameLen, vars->fileName) == 0)
   {
      vars->foundCatalog = loadCatalog(fileName, 0, vars->creator, vars->type,0);
      if (vars->foundCatalog)
         return 1;
   }
   return 0;
}

DmOpenRef myDmFindAndOpenDB(char *name, uint32 creator, uint32 type, byte readOnly, byte onCard)
{
   CatalogPtr db;
#ifndef PALMOS
   FileMatches_Vars vars;

   // 0. Does the filename specifies a path?
   if (name[0] == '\\' || name[0] == '.') // guich@400_57: allow the creation of pdbs in another directory
   {
      createFileName("", name, tempPath, ".pdb");
      db = loadCatalog(tempPath, 0, creator, type,0);
      if (db != NULL)
         return (DmOpenRef)db;
   }
#endif
   // 1. Search in the app path
   createFileName(getAppPath(), name, tempPath, ".pdb");
   db = loadCatalog(tempPath, 0, creator, type, readOnly);
   if (db != NULL)
      return (DmOpenRef)db;

#ifndef PALMOS // in palm, search only in the appPath
   // 2. Search in the sw path
   createFileName2(getSWPath(), name, tempPath, ".pdb");
   db = loadCatalog(tempPath, 0, creator, type, readOnly);
   if (db != NULL)
      return (DmOpenRef)db;

   // 3. Search in all SW program paths
   vars.fileNameLen = xstrlen(name)+4; // + .pdb
   vars.fileName = &tempPath[lstrlen(tempPath) - vars.fileNameLen];
   vars.creator = creator;
   vars.type = type;
   vars.foundCatalog = NULL;

   myDmListDatabases(&fileMatches, &vars, 1);

   return (DmOpenRef)vars.foundCatalog;
#else
   return 0;
#endif
}

/////////////////////////////////////////////////////////////////////////////
byte endsWithPDB(TCHAR *fName)
{
   int len = 0;
   while (*fName) {fName++; len++;}
   return (byte)(len > 3 && (*(fName-3) == (TCHAR)'p' || *(fName-3) == (TCHAR)'P')
                         && (*(fName-2) == (TCHAR)'d' || *(fName-2) == (TCHAR)'D')
                         && (*(fName-1) == (TCHAR)'b' || *(fName-1) == (TCHAR)'B'));
}
/////////////////////////////////////////////////////////////////////////////
char* myDmGetName(DmOpenRef dbref, char *buf)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (DB_IS_OPEN)
      xstrncpy(buf, db->dbh.name, xstrlen(db->dbh.name)+1);
   else
      *buf = 0;
   return buf;
}
/////////////////////////////////////////////////////////////////////////////
Err myDmCloseDatabase(DmOpenRef dbref)
{
   CatalogPtr db = (CatalogPtr)dbref;
   int wasOpen = DB_IS_OPEN;
   if (wasOpen) // write to disk
   {
      // close still opened records
      if (db->recordIndex != -1)
         myDmReleaseRecord(db, db->recordIndex, 1);
      // reset the open flag only if all records are unlocked
      if (db->lockedRecords <= 0)
         db->dbh.attributes &= ~dmHdrAttrOpen;
      else
      {
         (*vmGlobals->debug)("Records left locked in");
         (*vmGlobals->debug)("unprotected closed database:");
         (*vmGlobals->debug)(db->dbh.name);
         destroyRecPtrs(db);
      }

      // something has changed?
      if (db->originalModificationNumber != db->dbh.modificationNumber)
      {
         // set the platform dependent backup flag
//       SetFileAttributes(db->fh, GetFileAttributes(db->fh) | FILE_ATTRIBUTE_ARCHIVE); - guich@550: this cause problems in many platforms
         // update modification date
         db->dbh.modificationDate = _myGetNow();
         // write back the modified header
         swapDatabaseHeader(&db->dbh,&db->dbh);
         _myWriteAt(db->fh, &db->dbh, 78, 0);
      }
      else
      {
         db->dbh.attributes = READ_BE_UINT16(&db->dbh.attributes);
         #ifdef PALMOS
         _myWriteAt(db->fh, &db->dbh.attributes, 2, 32); // write only the attribute
         #endif
      }

      // close file and release memory
//    fseek(db->fh, 0, SEEK_END);
//    fflush(db->fh);
      _myCloseFile(db->fh);
      db->fh = NULL;
      if (db->recordBuf != NULL)
         xfree(db->recordBuf);
      xfree(db);
   }
   return wasOpen ? PALM_SUCCESS:PALM_ERROR;
}
/////////////////////////////////////////////////////////////////////////////
Err myDmGetDatabaseTypeCreatorCount(TCHAR *fullPath, uint32 *type, uint32 *creator, uint16 *recCount) // guich@340_50
{
   _HANDLE h = _myCreateFile(fullPath, 0, 1);    // exist, read_only - NOTE: IN PALM OS, THIS WAS DIFFERENT
   Err err = PALM_ERROR;
   *type = 0;
   *creator = 0;
   if (recCount) *recCount = 0;

   if (h)
   {
      err = _myReadAt(h, type, 4, 60);
      if (err == PALM_SUCCESS)
      {
         *type = READ_BE_UINT32(type);
         if (_myRead(h, creator, 4) == PALM_ERROR)
         {
            err = PALM_ERROR;
         }
         else
         {
            *creator = READ_BE_UINT32(creator);
            if (recCount)
            {
               err = _myReadAt(h, recCount, 2, 76);
               *recCount = READ_BE_UINT16(recCount);
            }
         }
      }
      _myCloseFile(h);
   }
   return err;
}
/////////////////////////////////////////////////////////////////////////////
Err myDmRenameDatabase(DmOpenRef dbref, uint32 newType, uint32 newCreator, char *newName)
{
   CatalogPtr db = (CatalogPtr)dbref;
   TCHAR tOldName[MAX_PATH], tNewName[MAX_PATH];
   int len;
   if (!DB_IS_OPEN || (len=xstrlen(newName)) > 31) return PALM_ERROR;
   // get the full path of the current file
   getDBFileName(db, tOldName);
   // change the internal fields
   xmemzero(db->dbh.name,sizeof(db->dbh.name)); // guich@400_37
   xmemzero(tNewName,sizeof(tNewName)); // guich@502_4: deleted the line that was erasing tOldName.
   xstrncpy(db->dbh.name,newName,len);
   db->dbh.creator = newCreator;
   db->dbh.type = newType;
   // get the full path to the new file
   getDBFileName(db, tNewName);
   // close the file
   myDmCloseDatabase(db);
   // physicaly rename the file
   _myRename(tOldName, tNewName);

   return 0;
}
/////////////////////////////////////////////////////////////////////////////
Err myDmDeleteDatabase(DmOpenRef dbref)
{
   CatalogPtr db = (CatalogPtr)dbref;
   TCHAR name[MAX_PATH];
   if (!DB_IS_OPEN) return PALM_ERROR;
   getDBFileName(db,name);
   myDmCloseDatabase(db);
   #ifndef __ECOS
   _myRemove(name);
   #endif
   return 0;
}
/////////////////////////////////////////////////////////////////////////////
int32 ldinPdbCount;
byte ldinStopSearch = 0;

int32 myDmListDatabases(HandlePDBSearchProcType proc, void *userVars, byte recursive)
{
   ldinPdbCount = 0;
   ldinStopSearch = 0;
   _myListDatabasesIn(getSWPath(), proc, userVars, recursive);
   return ldinPdbCount;
}
/////////////////////////////////////////////////////////////////////////////
int32 myDmListDatabasesIn(char * path, HandlePDBSearchProcType proc, void *userVars, byte recursive)
{
   ldinPdbCount = 0;
   ldinStopSearch = 0;
   ascii2unicode(tempPath, path, xstrlen(path)+1);
   _myListDatabasesIn(tempPath, proc, userVars, recursive);
   return ldinPdbCount;
}
/////////////////////////////////////////////////////////////////////////////
uint16 myDmNumRecords(DmOpenRef dbref)
{
   CatalogPtr db = (CatalogPtr)dbref;
   return db ? db->dbh.numRecords : 0;
}
/////////////////////////////////////////////////////////////////////////////
Err myDmWrite(DmOpenRef dbref, uint16 index, uint32 offset, const void *srcP, uint32 bytes)
{
   CatalogPtr db = (CatalogPtr)dbref;
   byte* recordP;
   uint16 size;
   RecordList *rl = NULL;
   if (!DB_IS_OPEN) return PALM_ERROR;
   if (index == db->recordIndex)
   {
      recordP = db->recordBuf;
      size = db->recordSize;
   }
   else
   {
      rl = retrieveRecList(db, index);
      if (rl == NULL) return PALM_ERROR;
      recordP = rl->recPtr;
      size = rl->size;
   }

   if ((offset+bytes) > size)
      return PALM_ERROR;
   xmemmove(((byte *)recordP)+offset, srcP, bytes);
   if (rl == NULL)
      db->recordChanged = 1;
   else
      rl->changed = 1;
   db->dbh.modificationNumber++;
   return PALM_SUCCESS;
}
/////////////////////////////////////////////////////////////////////////////
static int32 getRecordOffset(DmOpenRef dbref, int32 index)
{
   CatalogPtr db = (CatalogPtr)dbref;
   int32 offset;
   // get the offset to the current record index
   _myReadAt(db->fh, &offset, 4, GET_OFFSET(index));
   return READ_BE_INT32(&offset);
}
/////////////////////////////////////////////////////////////////////////////
static void shiftOffsets(DmOpenRef dbref, int32 start, int32 end, int32 delta)
{
   CatalogPtr db = (CatalogPtr)dbref;
   int32 count = end-start;
   int32 pos1 = GET_OFFSET(start);
   int32 i;

   while (count > 16)
   {
      _myReadAt(db->fh, rhs, 8*16, pos1);
      for (i = 0; i < 16; i++)
      {
         uint32 ofs = READ_BE_UINT32(&rhs[i].offset) + delta;   // swapRecordHeader
         rhs[i].offset = READ_BE_UINT32(&ofs);                  // swapRecordHeader
      }
      _myWriteAt(db->fh, rhs, 8*16, pos1);
      pos1 += 16*8;
      count -= 16;
   }
   if (count > 0)
   {
      _myReadAt(db->fh, rhs, 8*count, pos1);
      for (i = 0; i < count; i++)
      {
         uint32 ofs = READ_BE_UINT32(&rhs[i].offset) + delta;  // swapRecordHeader
         rhs[i].offset = READ_BE_UINT32(&ofs);                 // swapRecordHeader
      }
      _myWriteAt(db->fh, rhs, 8*count, pos1);
   }
}

/////////////////////////////////////////////////////////////////////////////
Err myDmReleaseRecord(DmOpenRef dbref, int32 index, byte dirty)
{
   CatalogPtr db = (CatalogPtr)dbref;
   byte attr=0;
   byte isNew;
   if (!DB_IS_OPEN)
      return PALM_ERROR;
   isNew = db->isNewRecord;
   if (!isNew && myDmGetRecordAttributes(db, (uint16)index, &attr) == PALM_ERROR)
      return PALM_ERROR;
   lastPos = lastPos2 = -1;
   if (isNew || (attr & dmRecAttrBusy))
   {
      int32 size,originalSize;
      RecordList *rl=NULL;
      void *buf;
      byte changed;
      if (index == db->recordIndex) // are we releasing the current record that is being written?
      {
         buf = db->recordBuf;
         size = db->recordSize;
         originalSize = db->originalRecordSize;
         changed = db->recordChanged;
      }
      else
      {
         rl = retrieveRecList(db, (uint16)index);
         if (rl == NULL)
            return PALM_ERROR;
         size = rl->size;
         originalSize = myDmGetRecordSize(db, (uint16)index);
         buf = rl->recPtr;
         changed = rl->changed;
      }

      if (buf == NULL) return PALM_ERROR;
      // if we changed the records size, we must resize the whole file
      if (size != originalSize || isNew)
      {
         byte appending=0;
         int32 pos1; // stores the position for this index in the record headers table
         int32 pos2; // stores the position where the record really starts.
         uint16 count = db->dbh.numRecords;
         int32 growingSize = isNew ? (size + ((db->fileSize==82)?6:8)) : (size - originalSize); // guich@550
         int32 delta = isNew ? size : growingSize;

         if (isNew) count--;
         if (index >= count)
         {
            index = count;
            appending = 1;
         }

         // 1. change the file size
         if (growingSize > 0)
         {
            if (_myGrowFileSize(db->fh, db->fileSize, growingSize) == PALM_ERROR)
               return PALM_ERROR;
            db->fileSize += growingSize;
         }

         pos1 = GET_OFFSET(index);

         // 2. if not appending, insert space (or remove space) for the new record, shifting the file to the right
         if (!appending)
         {
            // get the offset to the current record index
            _myReadAt(db->fh, &pos2, 4, pos1);
            pos2 = READ_BE_INT32(&pos2);
            // shift the file
            if (growingSize > 0)
            {
               if (isNew)
                  shiftFile(db, pos2, db->fileSize-growingSize, growingSize);
               else
                  shiftFile(db, pos2+originalSize, db->fileSize-growingSize, growingSize);
            }
            else
               shiftFile(db, pos2+originalSize, db->fileSize, growingSize);
         }
         else pos2 = db->fileSize - size;

         // 3. now we insert space for the new index at the offset table
         if (isNew)
            shiftFile(db, pos1, pos2, 8);

         // 4. write the record header for the new index
         {
            RecordHeader tmp;
            tmp.offset = READ_BE_UINT32(&pos2);                // swapRecordHeader
            tmp.attr = dmRecAttrDirty;
            tmp.uniqueId[0] = 0;
            tmp.uniqueId[1] = 0;
            tmp.uniqueId[2] = 0;
            _myWriteAt(db->fh, &tmp, 8, pos1);
         }

         // 5. now shift the offset of all records in the offsets table after this index
         if (isNew)
            shiftOffsets(db, 0,appending?index:(index+1),8); // since we inserted a new index, shift the record offsets 8 bytes down // guich@330_2: without the "appending?:", creating an empty catalog, adding one record and closing it, makes the first record have an offset of 96 instead of 88.
         if (!appending)
            shiftOffsets(db, index+1,isNew?(count+1):count,growingSize); // move the others down

         // 6. change the file size
         if (growingSize < 0)
         {
            if (_myGrowFileSize(db->fh, db->fileSize, growingSize) == PALM_ERROR)
               return PALM_ERROR;
            db->fileSize += growingSize;
         }

         if (db->dbh.appInfoOffset > 0)
            db->dbh.appInfoOffset += delta;
         if (db->dbh.sortInfoOffset > 0)
            db->dbh.sortInfoOffset += delta;
      }
      // write the record to disk
      if (changed && index >= 0)
         _myWriteAt(db->fh, buf, size, getRecordOffset(db, index));
      if (rl == NULL) // if not updating a record...
         db->recordIndex = -1;

      if (!isNew)
      {
         // unlock record
         attr &= ~dmRecAttrBusy; // clear busy attribute
         if (dirty)
            attr |= dmRecAttrDirty;
         myDmSetRecordAttributes(db, (uint16)index, attr);
         destroyRecPtr(db, (uint16)index);
      }
      else db->isNewRecord = 0;
      db->lockedRecords--;

      return PALM_SUCCESS;
   }
   return PALM_ERROR;
}

/////////////////////////////////////////////////////////////////////////////
static MemHandle resizeBuffer(DmOpenRef dbref, uint32 size)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (db->recordBuf == NULL || db->sizeofRecordBuf < size)
   {
      uint32 temp;
      temp = size;
#ifdef PALMOS
      if (temp < 1024) temp = 1024; // make sure that we allocate from the storage heap
#endif
      db->recordBuf = (byte*)vmGlobals->xrealloc(db->recordBuf, &temp);
      if (!db->recordBuf)
      {
         db->recordBuf = (byte*)vmGlobals->xmalloc(temp);
         temp = size;
      }
      if (temp >= size)
         db->sizeofRecordBuf = temp;
      else
      {
         xfree(db->recordBuf); // already set to null
         db->sizeofRecordBuf = 0;
      }
   }
   if (db->recordBuf)
   {
      db->dbh.modificationNumber++;
      db->recordSize = (uint16)size;
   }
   return (MemHandle)db->recordBuf;
}

/////////////////////////////////////////////////////////////////////////////
/** If the record is open, the MemHandle to it will be returned. If the record is closed,
it will be opened, but a non-zero (but invalid) value will be returned */
MemHandle myDmResizeRecord(DmOpenRef dbref, uint16 index, uint32 size)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (!DB_IS_OPEN || size == 0 || size > 65520L)
      return NULL; // closed, invalid index or invalid size?
   if (db->recordIndex != index) // resizing a record got with getRecord? (updating a record?)
   {
      RecordList *rl = retrieveRecList(db,index);
      uint32 temp = size;
      MemHandle handleToBeReleased = NULL;
      if (rl == NULL) // guich@552_25: if the record was not loaded, load and then release it instead of just returning an error
      {
         handleToBeReleased = myDmGetRecord(dbref, index);
         if (!handleToBeReleased)
            return NULL;
         rl = retrieveRecList(db,index);
         if (rl == NULL) // if failed for the second time, we have no options beside just return
            return NULL;
      }
      if (rl->size != size)
      {
         rl->recPtr = (byte*)vmGlobals->xrealloc(rl->recPtr, &temp);
         if (temp < size) // guich@550_@@: changed != to <  bc i had changed the realloc routine
            return NULL;
         rl->changed = 1;
         rl->size = (uint16)size;
      }
      db->dbh.modificationNumber++;
      if (handleToBeReleased) // guich@552_25
      {
         myDmReleaseRecord(dbref, index, 1);
         return (MemHandle)1; // just to tell that the record was resized with success
      }
      return (MemHandle)rl->recPtr;
   }
   else return resizeBuffer(db, size);
}

/////////////////////////////////////////////////////////////////////////////
/* Just create memory for the record; the record will be written and
the file will be shifted in myDmReleaseRecord. */
MemHandle myDmNewRecord(DmOpenRef dbref, uint16 *atP, uint32 size)
{
   CatalogPtr db = (CatalogPtr)dbref;
   uint16 numRecords;
   if (!DB_IS_OPEN || size == 0 || size > 65520L) return NULL; // closed or invalid size?
   if (db->isNewRecord)
   {
      (*vmGlobals->debug)("Already adding a record. Please call DmReleaseRecord before adding another one");
      return NULL;
   }

   // 0. create the buffer that will store the data - used only in myDmWrite
   if (resizeBuffer(db, size) == NULL)
      return NULL;
   xmemzero(db->recordBuf,size);

   // 1. recompute the internal parameters
   if (*atP == 65535) // guich@360_94
      *atP = db->dbh.numRecords;
   db->recordIndex = *atP;
   db->isNewRecord = 1;  // when releasing the record, shift the file
   db->recordChanged = 0;
   db->dbh.numRecords++;
   db->originalRecordSize = (uint16)size;
   db->lockedRecords++;

   // update the number of records
   numRecords = READ_BE_UINT16(&db->dbh.numRecords);
   _myWriteAt(db->fh, &numRecords, 2, 76);
   return (MemHandle)db->recordBuf;
}

/////////////////////////////////////////////////////////////////////////////
Err myDmRemoveRecord(DmOpenRef dbref, uint16 index)
{
   CatalogPtr db = (CatalogPtr)dbref;
   uint16 numRecords;
   byte appending = 0; // well, not really appending, but removing from end (how to say this in english? :-) )
   int32 pos1; // stores the position for this index in the record headers table
   int32 pos2; // stores the position where the record really starts.
   int32 shrinkingSize;
   uint16 size;

   if (!DB_IS_OPEN || index >= db->dbh.numRecords || db->recordIndex != -1) return PALM_ERROR; // closed or invalid size?

   size = (uint16)myDmGetRecordSize(db, index);

   if (index == (db->dbh.numRecords-1))
      appending = 1;

   pos1 = GET_OFFSET(index);
   shrinkingSize = - ((int32)size + 8);

   // 2. if not appending, remove the space of the current record, shifting the file to the right
   if (!appending)
   {
      _myReadAt(db->fh, &pos2, 4, pos1);
      pos2 = READ_BE_INT32(&pos2);
   }
   else
      pos2 = db->fileSize-size;

   // 3. records before this one only shift by the removed entry of the records table
   shiftOffsets(db, 0, index, -8);

   // 4. records after this one shifts the total bytes removed
   shiftOffsets(db, index+1, db->dbh.numRecords, shrinkingSize);

   // 5. now we remove the space of the index at the offset table
   shiftFile(db, pos1+8, pos2, -8);

   // 6. and now we remove the space of the record
   if (!appending)
      shiftFile(db, pos2+size, db->fileSize, shrinkingSize); // shrinkingSize is always < 0

   // 7. change the file size
   if (_myGrowFileSize(db->fh, db->fileSize, shrinkingSize) == PALM_ERROR)
      return PALM_ERROR;
   db->fileSize += shrinkingSize;

   // 8. recompute the internal parameters
   db->dbh.numRecords--;
   db->dbh.modificationNumber++;
   db->recordChanged = 0;
   if (db->dbh.appInfoOffset > 0)
      db->dbh.appInfoOffset += size;
   if (db->dbh.sortInfoOffset > 0)
      db->dbh.sortInfoOffset += size;
   // update the number of records
   numRecords = READ_BE_UINT16(&db->dbh.numRecords);
   _myWriteAt(db->fh, &numRecords, 2, 76);
   return PALM_SUCCESS;
}

/////////////////////////////////////////////////////////////////////////////
int32 myDmGetRecordSize(DmOpenRef dbref, uint16 index)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (!DB_IS_OPEN || (index >= db->dbh.numRecords))
   {
      return 0;
   }
   else
   {
      static int32 size = 0;
      static RecordHeader rh;
      static int32 lastfh = -1;

      if ((int32)db->fh != lastfh || index != lastPos2) // guich@550: do a little cache
      {
         _myReadAt(db->fh, &rh, 8, GET_OFFSET(index));  // get current record header
         rh.offset = READ_BE_UINT32(&rh.offset);         // swapRecordHeader
         if (index < (db->dbh.numRecords-1))
         {
            _myRead(db->fh, &size, 4);             // get next offset
            size = READ_BE_INT32(&size);
         }
         else
         {
            size = db->fileSize;
         }
         size -= rh.offset; // sub
         size = (size < 0) ? 0 : size;
         lastPos2 = index;
         lastfh = (int32)db->fh;
      }
      return size;
   }
}

/////////////////////////////////////////////////////////////////////////////
static MemHandle getRecord(DmOpenRef dbref, uint16 index, int32 *size, byte queryOnly, int32 ofs, byte* buf)
{
   CatalogPtr db = (CatalogPtr)dbref;
   int32 offset = GET_OFFSET(index); // offset to the offsets table
   static int32 realSize;
   static RecordHeader rh;
   static int32 lastfh = -1;
   byte* m=0;

   if (!DB_IS_OPEN || index >= db->dbh.numRecords) return NULL;

   if ((int32)db->fh != lastfh || index != lastPos) // guich@550: do a little cache
   {
      _myReadAt(db->fh, &rh, 8, offset);       // get current record header
      rh.offset = READ_BE_UINT32(&rh.offset);    // swapRecordHeader

      // get record offset and real size
      if (index < (db->dbh.numRecords-1))
      {
         _myRead(db->fh, &realSize, 4); // get next offset
         realSize = READ_BE_INT32(&realSize);
      }
      else
      {
         realSize = db->fileSize;
      }
      realSize -= rh.offset; // sub
      if (realSize < 0) // error?
      {
         return 0;
      }
      lastPos = index;
      lastfh = (int32)db->fh;
   }

   if (!queryOnly && (rh.attr & dmRecAttrBusy)) // is the record already locked?
      return 0;
   if (size == 0)
      size = &realSize;
   else if (*size == 0)
      *size = realSize;  // just want to get the record size
   else
      *size = (realSize < *size)?realSize:*size; // wants to read a piece of the record only
   if (!buf)
   {
      m = (byte*)(*vmGlobals->xmalloc)(*size);
      if (!m) return 0;
   }
   _myReadAt(db->fh, !buf?m:buf, *size, rh.offset+ofs); // can't change to READ_SIZE_AT bc the macro uses & in the pointer, corrupting memory

   // set the busy bit
   if (!queryOnly)
   {
      if (!storeRecPtr(db, m, index, (uint16)*size))
      {
         xfree(m);
      }
      else
      {
         rh.attr |= dmRecAttrBusy;
         _myWriteAt(db->fh, &rh.attr, 1, offset+4);
         db->originalRecordSize = (uint16)realSize;
         db->recordChanged = 0;
         db->lockedRecords++;
      }
   }
   return (MemHandle)(!buf?m:buf);
}

/////////////////////////////////////////////////////////////////////////////
MemHandle myDmGetRecord(DmOpenRef dbref, uint16 index)
{
   CatalogPtr db = (CatalogPtr)dbref;
   return getRecord(db, index, NULL, 0,0,0);
}
/////////////////////////////////////////////////////////////////////////////
MemHandle myDmQueryRecord(DmOpenRef dbref, uint16 index, int32 *size)
{
   CatalogPtr db = (CatalogPtr)dbref;
   return getRecord(db, index, size, 1,0,0);
}
/////////////////////////////////////////////////////////////////////////////
Err myDmQueryRecordPart(DmOpenRef dbref, uint16 index, int32 size, int32 ofs, byte* buf) // guich@300
{
   CatalogPtr db = (CatalogPtr)dbref;
   return (getRecord(db, index, &size, 1, ofs, buf) != NULL) ? PALM_SUCCESS : PALM_ERROR;
}
/////////////////////////////////////////////////////////////////////////////
Err myDmGetRecordAttributes(DmOpenRef dbref, uint16 index, byte *attrs)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (!DB_IS_OPEN || index >= db->dbh.numRecords) return PALM_ERROR;
   return _myReadAt(db->fh, attrs, 1, GET_OFFSET(index)+4);
}
/////////////////////////////////////////////////////////////////////////////
Err myDmSetRecordAttributes(DmOpenRef dbref, uint16 index, byte attrs)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (!DB_IS_OPEN || index >= db->dbh.numRecords) return PALM_ERROR;
   ++db->dbh.modificationNumber;
   return _myWriteAt(db->fh, &attrs, 1, GET_OFFSET(index)+4);
}
/////////////////////////////////////////////////////////////////////////////
Err myDmSetAttributes(DmOpenRef dbref, int16 attrs)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (!DB_IS_OPEN) return PALM_ERROR;
   db->dbh.attributes = attrs;
   db->dbh.modificationNumber++;
   return PALM_SUCCESS;
}

/////////////////////////////////////////////////////////////////////////////
Err myDmGetAttributes(DmOpenRef dbref, int16 *attrs)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (!DB_IS_OPEN) return PALM_ERROR;
   *attrs = db->dbh.attributes;
   return PALM_SUCCESS;
}

/////////////////////////////////////////////////////////////////////////////
Err myDmSetBackupAttribute(DmOpenRef dbref)
{
   CatalogPtr db = (CatalogPtr)dbref;
   if (!DB_IS_OPEN) return PALM_ERROR;
   return myDmSetAttributes(db, (int16)(db->dbh.attributes | dmHdrAttrBackup));
}

/*-------------------------------------------------------------setPalmDBProcs-+
|                                                                             |
+----------------------------------------------------------------------------*/
void setPalmDBProcs(struct VMGlobals * vmGlobals)
{
#ifndef PALMOS
   vmGlobals->_splitPath = mySplitPath;
   vmGlobals->findAndOpenDB = myDmFindAndOpenDB;
   vmGlobals->_DmCloseDatabase = myDmCloseDatabase;
   vmGlobals->_DmCreateDatabase = myDmCreateDatabase;
   vmGlobals->_DmOpenDatabase = myDmOpenDatabase;
   vmGlobals->_DmOpenDatabaseReadOnly = myDmOpenDatabaseReadOnly;
   vmGlobals->_DmRenameDatabase = myDmRenameDatabase;
   vmGlobals->_DmDeleteDatabase = myDmDeleteDatabase;
   vmGlobals->_DmListDatabases = myDmListDatabases;
   vmGlobals->_DmListDatabasesIn = myDmListDatabasesIn;
   vmGlobals->_DmNumRecords = myDmNumRecords;
   vmGlobals->_DmWrite = myDmWrite;
   vmGlobals->_DmReleaseRecord = myDmReleaseRecord;
   vmGlobals->_DmNewRecord = myDmNewRecord;
   vmGlobals->_DmResizeRecord = myDmResizeRecord;
   vmGlobals->_DmRemoveRecord = myDmRemoveRecord;
   vmGlobals->_DmGetRecord = myDmGetRecord;
   vmGlobals->_DmQueryRecord = myDmQueryRecord;
   vmGlobals->_DmGetRecordAttributes = myDmGetRecordAttributes;
   vmGlobals->_DmSetRecordAttributes = myDmSetRecordAttributes;
   vmGlobals->_DmSetAttributes = myDmSetAttributes;
   vmGlobals->_DmGetAttributes = myDmGetAttributes;
   vmGlobals->_DmSetBackupAttribute = myDmSetBackupAttribute;
   vmGlobals->_DmGetRecordSize = myDmGetRecordSize;
   vmGlobals->_DmGetDatabaseTypeCreatorCount = myDmGetDatabaseTypeCreatorCount;
   vmGlobals->_DmGetName = myDmGetName;
   vmGlobals->_DmQueryRecordPart = myDmQueryRecordPart;
#endif
}

/*===========================================================================*/
