/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_settings_posix.c,v 1.11 2005/09/24 21:25:38 guich Exp $ */

#include <locale.h>
#include "superwaba.h"

#if HAVE_CONFIG_H
#include "config.h"
#endif

#define __DEFINE_FUNCTIONS
#include "SW_String.h"
#include "SW_utils.h"
#include "SW_settings.h"

#if HAVE_STRING_H
#include <string.h>
#endif

#include <time.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <sys/fcntl.h>

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

// from superwaba.c:
WClass * getClass(UtfString className, byte throwEx);

#define HIVE_MAXCHARS      ((4 + sizeof(uint32) * 8) / 5)
#define BLK_SZ             128
#define DEFAULT_DIR_PERMS  0775
#define DEFAULT_FILE_PERMS 0664

typedef TCHAR AppHive[HIVE_MAXCHARS + 2];  // +1 for the dot & '\0'

static const TCHAR *settingsSubdir = TEXT("appSettings/");

static void getAppPreferencesPath(TCHAR *buf, AppHive h)
{
   lstrcpy(buf, vmGlobals->swPath);
   lstrcat(buf, settingsSubdir);
   lstrcat(buf, h);
}

static int ensureAppPreferencesDirectory(TCHAR * buf)
{
   lstrcpy(buf, vmGlobals->swPath);
   if (access(buf, W_OK) != 0) 
   {
      if (errno != ENOENT) return 0;
      if (mkdir(buf, DEFAULT_DIR_PERMS) != 0) return 0;
   }
   lstrcat(buf, settingsSubdir);
   if (access(buf, W_OK) != 0)
   {
      if (errno != ENOENT) return 0;
      if (mkdir(buf, DEFAULT_DIR_PERMS) != 0) return 0;
   }
   return 1;
}

static TCHAR *getAppHive(AppHive h, uint32 crid)
{
   char *cur = h;
   *cur++ = (TCHAR)      (crid >> 24);
   *cur++ = (TCHAR)LOWER8(crid >> 16);
   *cur++ = (TCHAR)LOWER8(crid >> 8);
   *cur++ = (TCHAR)LOWER8(crid);
   *cur++ = 0;
   return h;
}

static TCHAR *getAppSecretHive(AppHive h, uint32 crid)
{
   TCHAR *buf = h;
   unsigned i;
   
   *buf++ = '.';
   // This is not base32!  sure :-D
   for (i = 0; i < HIVE_MAXCHARS; i++)
   {
      char c = crid & 0x1F;
      crid >>= 5;
      c += 'A';
      if (c > 'Z') c += ('2' - 'Z' - 1);
      *buf++ = c;
   }
   *buf = '\0';
   return h;
}

static char *readAppPreferences(TCHAR *hive)
{
   int f;
   TCHAR buf[MAX_PATH];
   getAppPreferencesPath(buf, hive);

   _LOG_INFO((LOG_CAT_INIT, "readAppPreferences: %s", buf));
   if ((f = open(buf, O_RDONLY)) >= 0)
   {
      char *pref = 0;
      char temp[BLK_SZ];
      uint32 total = 0;
      uint32 count;
      while ((count = read(f, temp, sizeof temp)) > 0)
      {
         int i = total + count + 1;
         pref = xrealloc(pref, &i);
         if (!pref) break;
         memcpy(pref+total, temp, count);
         total += count;
         pref[total] = 0;
      }
      close(f);
      return pref;
   }
   else
   {
      _LOG_DEBUG((LOG_CAT_INIT, "readAppPreferences failed: %d", errno));
   }
   return 0;
}

static void writeAppPreferences(AppHive h, UtfString utf)
{
   TCHAR buf[MAX_PATH];

   _LOG_INFO((LOG_CAT_INIT, "writeAppPreferences: hive=%s", h));
   if (ensureAppPreferencesDirectory(buf))
   {
      int f;
      getAppPreferencesPath(buf, h);
      _LOG_DEBUG((LOG_CAT_INIT, "writeAppPreferences: %s", buf));
      if ((f = open(buf, O_CREAT|O_WRONLY, 0600)) >= 0)
      {
         write(f, utf.str, utf.len);
         close(f);
      }
      else
      {
         _LOG_ERROR((LOG_CAT_INIT, "writeAppPreferences failed: %d", errno));
      }
   }
}

static void deleteAppPreferences(AppHive h)
{
   TCHAR buf[MAX_PATH];
   getAppPreferencesPath(buf, h);
   unlink(buf);
}

// frank@500_01: class member checking
static Var * getSettingsField(char * name, char * desc)
{
   UtfString fieldName = createUtfString(name);
   UtfString fieldDesc = createUtfString(desc);
   WClassField * field = vmGlobals->getField(vmGlobals->prefClass, &fieldName, &fieldDesc);
   if (field)
   {
      return &field->var.staticVar;
   }
   return 0;
}

static void setSettingsIntField(char *name, char *desc, int32 value)
{
   Var * var = getSettingsField(name, desc);
   if (var)
   {
      var->intValue = value;
   }
}

static void setSettingsObjField(char *name, char *desc, WObject obj)
{
   Var * var = getSettingsField(name, desc);
   if (var)
   {
      var->obj = obj;
   }
}

static void storeSettingsEntry(char *name, AppHive h)
{
   Var * var = getSettingsField(name, "Ljava/lang/String;");
   if (var)
   {
      WObject obj = var->obj;
      if (obj == NULL || WOBJ_StringCharArrayObj(obj) == NULL)
      {
         // if string null, delete it - guich@240_3: first condition added.
         deleteAppPreferences(h);
      }
      else
      {
         writeAppPreferences(h, stringToUtf(obj, STU_USE_STATIC | STU_NULL_TERMINATE));
      }
   }
}

/*----------------------------------------------------------------getSettings-+
|                                                                             |
+----------------------------------------------------------------------------*/
void getSettings()
{
   char tmpbuf[10];
   char *pref;
   char ts;                 // decimal separator
   char ds;                 // thousands separator
   char dateSep = '/';      // datefmt separator
   char timeSep = ':';      // time separator
   int weekStart = 0;       // week start
   int time24h;             // time 0-24h
   int datefmt;             // date format
   int gmtBias;             // gmt+0
   int daylightSavings;     // 1 when DST is on
   int haveVirtualKeyboard;
#if defined(__SYMBIAN32__) && NO_PEN
   int penless = 1;
#else
   int penless = 0;
#endif   
   int32 romVersion = 0;
   AppHive hive;

   WObject userName;
   WObject platform;
   WObject appSettings = 0;
   WObject appCreatorId = 0;
   WObject appSecretKey = 0;
   WObject romSerialNumber = 0;

   time_t t;
   struct tm * tm;
   struct lconv * locale;

   // #ifndef NDEBUG
   //   vmGlobals->audibleGC = 1;
   //   vmGlobals->dumpMemStats = 1;
   // #endif

   #if HAVE_GETENV
   {
      char *user = getenv("USER");
      userName = createString(user ? user:"noname");
   }
   #else
      userName = createString("noname");
   #endif

   haveVirtualKeyboard = 0;           // the SIP is not implemented
   #if defined __SYMBIAN32__
      platform = createString("Symbian");
      if (!penless) haveVirtualKeyboard = 1; // let the keyboard appear on P800
   #elif defined linux
      platform = createString("Linux");
   #else
      platform = createString("Posix");  // String platform
   #endif

   #if defined __SYMBIAN32__ // fdie@502_12
      epoc_settings_get(&ds,&ts,&dateSep,&timeSep,&datefmt,&time24h,&daylightSavings,&weekStart,&gmtBias);
   #else
      // obtain number formats
      locale = localeconv();
      ts = locale->thousands_sep[0];
      ds = locale->decimal_point[0];
      if (ts == '\0') ts = ',';  // so poor microsoft!

      // obtain the time locale data
      time(&t);
      daylightSavings = localtime(&t)->tm_isdst;
      t = 30326400;   // 12/17/70 00:00:00 - GMT
      tm = localtime(&t);
      gmtBias = (tm->tm_hour * 60) + tm->tm_min;
      if ((gmtBias >= 720) && (tm->tm_mday == 17))
      {
        gmtBias -= 1440;
      }
      strftime(tmpbuf, sizeof tmpbuf, "%x", tm);
      if (strlen(tmpbuf) == 8)
      {
         dateSep = tmpbuf[2];
         if (tmpbuf[1] == '0')
         {
            datefmt = 3;                // 70/12/17
         }
         else if (tmpbuf[1] == '7')
         {
            datefmt = 2;                // 17/12/70
         }
         else
         {
            datefmt = 1;                // 12/17/70
         }
      }
      else 
      {
        datefmt = 0; // ?
      }
      tm->tm_hour = 22;
      strftime(tmpbuf, sizeof tmpbuf, "%X", tm);
      if (strlen(tmpbuf) == 8)
      {
         timeSep = tmpbuf[2];
      }
      time24h = (tmpbuf[0] > '1');
   #endif

   if ((pref = readAppPreferences(getAppHive(hive, vmGlobals->appCreatorId))) != NULL)
   {
      appSettings = (*vmGlobals->createString)(pref);
      xfree(pref);
   }

   // guich@330_47: get the size then the app secret key, if any
   if ((pref = readAppPreferences(getAppSecretHive(hive, vmGlobals->appCreatorId))) != NULL)
   {
      appSecretKey = (*vmGlobals->createString)(pref);
      xfree(pref);
   }

   // guich@230_22: get the creator id
   tmpbuf[0] = (char)((vmGlobals->appCreatorId >> 24) & 0xFF);
   tmpbuf[1] = (char)((vmGlobals->appCreatorId >> 16) & 0xFF);
   tmpbuf[2] = (char)((vmGlobals->appCreatorId >> 8 ) & 0xFF);
   tmpbuf[3] = (char)((vmGlobals->appCreatorId      ) & 0xFF);
   tmpbuf[4] = 0;
   appCreatorId = (*vmGlobals->createString)(tmpbuf);

   // store the preferences in the waba.sys.Settings class
   vmGlobals->prefClass = getClass(createUtfString("waba/sys/Settings"),1);

   // guich@421_17: don't call the constructor; just set directly the values
   // frank@500_01: don't crash if the member doesn't exist
   setSettingsIntField("dateFormat", "B", datefmt);
   setSettingsIntField("dateSeparator", "C", dateSep);
   setSettingsIntField("weekStart", "B", weekStart);
   setSettingsIntField("is24Hour", "Z", time24h);
   setSettingsIntField("timeSeparator", "C", timeSep);
   setSettingsIntField("thousandsSeparator", "C", ts);
   setSettingsIntField("decimalSeparator", "C", ds);
   setSettingsIntField("screenWidth", "I", vmGlobals->screenWidth);
   setSettingsIntField("screenHeight", "I", vmGlobals->screenHeight);
   setSettingsIntField("isColor", "Z", (vmGlobals->maxColors == 0) || (vmGlobals->maxColors > 16));
   setSettingsIntField("maxColors", "I", vmGlobals->maxColors);
   setSettingsIntField("onDevice", "Z", 1);
   setSettingsIntField("penless", "Z", penless);
   setSettingsIntField("romVersion", "I", romVersion);
   setSettingsObjField("platform", "Ljava/lang/String;", platform);
   setSettingsObjField("userName", "Ljava/lang/String;", userName);
   setSettingsObjField("appSettings", "Ljava/lang/String;", appSettings);
   setSettingsObjField("appCreatorId", "Ljava/lang/String;", appCreatorId);
   setSettingsIntField("virtualKeyboard", "Z", haveVirtualKeyboard);
   setSettingsObjField("appSecretKey", "Ljava/lang/String;", appSecretKey);
   setSettingsIntField("daylightSavings", "Z", daylightSavings);
   setSettingsIntField("timeZone", "I", gmtBias);
   setSettingsObjField("romSerialNumber", "Ljava/lang/String;", romSerialNumber);
   setSettingsIntField("isHighColor", "Z", (vmGlobals->maxColors > 256) || (vmGlobals->screenBPP > 8)); // guich@421_35
}

/*--------------------------------------------------------------storeSettings-+
|                                                                             |
+----------------------------------------------------------------------------*/
void storeSettings() // guich@230_22
{
   AppHive hive;
   storeSettingsEntry("appSettings", getAppHive(hive, vmGlobals->appCreatorId));
   storeSettingsEntry("appSecretKey", getAppSecretHive(hive, vmGlobals->appCreatorId)); // guich@330_47
}

/*--------------------------------------------------------------------showSip-+
| show, hide or move the sip                                                  |
+----------------------------------------------------------------------------*/
void showSip(int16 on, int16 ontop)
{
}

/*===========================================================================*/
