/* $Id: AudioServer.cpp,v 1.5 2004/09/22 10:38:26 petrus Exp $ */

#include "AudioServer.h"
#include "AudioTone.h"
#include "AudioPlayer.h"
#include <e32svr.h>
#include <e32math.h>
#include <e32uid.h>

TUint const heapSize=0x10000;

// panic reasons
enum AudioServerPanic {
   EBadRequest,
   EBadDescriptor,
   EMainSchedulerError,
   ESvrCreateServer,
   ECreateTrapCleanup
};

/*---------------------------------------------------------------------E32Dll-+
| DLL entry point                                                             |
+----------------------------------------------------------------------------*/
GLDEF_C TInt E32Dll(TDllReason /*reason*/)
{
   return KErrNone;
}

/*-------------------------------------------------------- class AudioServer -+
|                                                                             |
+----------------------------------------------------------------------------*/
class AudioServer : public CServer
{
public:
   static AudioServer * NewL() {
      AudioServer* self = new (ELeave) AudioServer;
      self->StartL(KAudioServerName);
      return self;
   }

   AudioServer() : CServer(/*EPriorityRealTime*/EPriorityAbsoluteForeground) {
      m_semAudioIdle.CreateGlobal(KAudioServerName, 0);
   }

   virtual ~AudioServer() {
      m_semAudioIdle.Close();
   }

   CSharableSession * NewSessionL(const TVersion & version) const;
   static TInt startServer(TAny * started);

   RSemaphore m_semAudioIdle;
};

/*---------------------------------------------------AudioServer::startServer-+
|                                                                             |
+----------------------------------------------------------------------------*/
TInt AudioServer::startServer(TAny * started)
{
   __UHEAP_MARK;

   CTrapCleanup * cleanup = CTrapCleanup::New();
   CServer * server = NULL;

   if (!cleanup) {
      User::Panic(KAudioServerName, ECreateTrapCleanup);
   }
   RSemaphore & semStarted = *(RSemaphore *)started; // cast to semaphore

   // construct and install active scheduler
   CActiveScheduler* scheduler = new CActiveScheduler;
   if (!scheduler) {
      User::Panic(KAudioServerName, EMainSchedulerError);
   }
   CActiveScheduler::Install(scheduler);

   // construct server, an active object
   TRAPD(err, server=AudioServer::NewL());
   if (err != KErrNone) {
      User::Panic(KAudioServerName, ESvrCreateServer);
   }
   semStarted.Signal();          // signal we started
   CActiveScheduler::Start();    // start the RunL loop

   // arrives here when either the scheduler is stopped, or RunL errored
   delete server;                // scheduler stopped: delete!
   delete CActiveScheduler::Current();
   delete cleanup;

   __UHEAP_MARKEND;
   return KErrNone;
}

/*-----------------------------------------------------------startAudioServer-+
| Called by the Client to start the audio server.                             |
+----------------------------------------------------------------------------*/
EXPORT_C TInt startAudioServer()
{
   // check server not already started
   TFindServer findAudioServer(KAudioServerName);
   TFullName name;
   if (findAudioServer.Next(name) == KErrNone) {
      return KErrAlreadyExists;
   }

   // create temporary thread
   RThread thread;
   RSemaphore semStarted;
   semStarted.CreateLocal(0);
   TInt err=thread.Create(
      KAudioServerName,
      AudioServer::startServer,
      KDefaultStackSize,
      heapSize,
      heapSize,
      &semStarted // parameter to thread function
   );

   if (err != KErrNone)
   {
      thread.Close();
      return err;
   }

   // run the thread to start the server
   thread.SetPriority(EPriorityMuchMore);
   thread.Resume();       // start running
   semStarted.Wait();     // wait until done
   thread.Close();        // clean up
   semStarted.Close();

   #if defined (__WINS__)
   UserSvr::ServerStarted(); // notify kernel that a server has started.
   #endif
   return KErrNone;
}

/*------------------------------------------------------- class AudioSession -+
|                                                                             |
+----------------------------------------------------------------------------*/
class AudioSession : public CSession
{
public:
   //construct/destruct/close session
   static AudioSession * NewL(
      RThread & client,
      /*const */ AudioServer * server
   );
   void ServiceL(const RMessage & message);
   void endServer();

private:
   AudioServer * m_audioServer; // pointer to owning server
   AudioSession(RThread & client);
   void ConstructL(AudioServer * server);
   void closeSession();
   void dispatchMessage(const RMessage & message);
   void panicClient(TInt panicCode) const;
};

/*---------------------------------------------------AudioServer::NewSessionL-+
| Create a new client session for this server.                                |
+----------------------------------------------------------------------------*/
CSharableSession * AudioServer::NewSessionL(
   TVersion const & version
) const {
   // check version is ok
   TVersion v(
      KAudioServerMajorVersionNumber,
      KAudioServerMinorVersionNumber,
      KAudioServerBuildVersionNumber
   );
   if (!User::QueryVersionSupported(v, version)) {
      User::Leave(KErrNotSupported);
   }
   // make new session
   RThread client= Message().Client();
   return AudioSession::NewL(client, (AudioServer *)this);
}

/*---------------------------------------------------------AudioSession::NewL-+
|                                                                             |
+----------------------------------------------------------------------------*/
AudioSession* AudioSession::NewL(RThread & client, AudioServer * server)
{
   AudioSession * self = new (ELeave) AudioSession(client);
   CleanupStack::PushL(self);
   self->ConstructL(server);
   CleanupStack::Pop();
   return self;
}


/*-------------------------------------------------AudioSession::AudioSession-+
|                                                                             |
+----------------------------------------------------------------------------*/
AudioSession::AudioSession(RThread & client) : CSession(client)
{
}


/*---------------------------------------------------AudioSession::ConstructL-+
| 2nd phase constructor                                                       |
+----------------------------------------------------------------------------*/
void AudioSession::ConstructL(AudioServer * server)
{
   CSession::CreateL(*server);      // construct base class
   m_audioServer = server;
}


/*----------------------------------------------------AudioSession::endServer-+
|                                                                             |
+----------------------------------------------------------------------------*/
void AudioSession::endServer()
{
   CActiveScheduler::Stop();
}

/*-------------------------------------------------AudioSession::closeSession-+
|                                                                             |
+----------------------------------------------------------------------------*/
void AudioSession::closeSession()
{
}

/*-----------------------------------------------------AudioSession::ServiceL-+
| Handle messages send by client(s)                                           |
+----------------------------------------------------------------------------*/
void AudioSession::ServiceL(const RMessage & message)
{
   TRAPD(err, dispatchMessage(message));
   message.Complete(err);
}

/*----------------------------------------------AudioSession::dispatchMessage-+
|                                                                             |
+----------------------------------------------------------------------------*/
void AudioSession::dispatchMessage(const RMessage & message)
{
   RMessage const & msg = Message();

   switch (message.Function()) {

   case EAudioServerEnable:
      // enable(msg.Int(0));
      m_audioServer->m_semAudioIdle.Signal();
      break;

   case EAudioServerPlayTone:
      new AudioTone(
         msg.Int0(),
         msg.Int1(),
         m_audioServer->m_semAudioIdle
      );
      break;

   case EAudioServerMakeBufHandle:
      {
         TAny * player = new AudioPlayer(
            (char *)msg.Ptr0(),
            m_audioServer->m_semAudioIdle
         );
         TPckg<TAny*> pkgHandle(player);
         WriteL(Message().Ptr1(), pkgHandle);
      }
      break;

   case EAudioServerMakeFileHandle:
      {
         TAny * player = new AudioPlayer(
            (unsigned short *)msg.Ptr0(),
            msg.Int1(),
            m_audioServer->m_semAudioIdle
         );
         TPckg<TAny*> pkgHandle(player);
         WriteL(Message().Ptr1(), pkgHandle);
      }
      break;

   case EAudioServerPlayHandle:
      ((AudioPlayer *)msg.Ptr0())->play();
      break;

   case EAudioServerFreeHandle:
      delete ((AudioPlayer *)msg.Ptr0());
      break;

   case EAudioServerCloseSession:
      closeSession();
      m_audioServer->m_semAudioIdle.Signal();
      break;

   case EAudioServerEndServer:
      endServer();
      m_audioServer->m_semAudioIdle.Signal();
      break;

   default:
      panicClient(EBadRequest);
      m_audioServer->m_semAudioIdle.Signal();
      break;
   }
}

/*--------------------------------------------------AudioSession::panicClient-+
|                                                                             |
+----------------------------------------------------------------------------*/
void AudioSession::panicClient(TInt panicCode) const
{
   _LIT(KTxtAudioServerSess,"AudioSession");
   Panic(KTxtAudioServerSess, panicCode);
}

/*===========================================================================*/
