/*********************************************************************************
 * Permission to use, copy, modify, and distribute, this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holders be used in
 * advertising or publicity pertaining to distribution of the software with
 * specific, written prior permission, and that no fee is charged for
 * further distribution of this software, or any modifications thereof. The
 * copyright holder make no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 * 
 * THE COPYRIGHT HOLDER DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA, PROFITS, QPA OR GPA, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *********************************************************************************/
/* $Id$ */

//Author Nataly Merenbach
//23.02.03

#include <e32base.h>
#include <e32cons.h>
#include <e32svr.h>
#include <e32math.h>
#include <e32uid.h>
#include <estlib.h>

#include "epoc_stat.h"

enum EStaticServerFunctions
{
	EGetStatics = 1
};

// ================ client ===========

class CStaticRepository
	: public RSessionBase
{

public:

	CStaticRepository() {}
	TAny* GetStatics();
	TInt Connect( const TDesC& aName,
			const StaticCreateFunction aStatFun);

private:

	struct SThreadInitData
	{
		TRequestStatus* status;
		TThreadId tid;
		const TDesC& server_name;
		const StaticCreateFunction statFun;

		SThreadInitData( const TDesC& aName,
				const StaticCreateFunction aStatFun)
			: server_name( aName), statFun( aStatFun)
		{}
	};

	static TInt ThreadFunction( TAny* anArg);
	static TInt StartServerThread( const TDesC& aName,
			const StaticCreateFunction aStatFun);

	CStaticRepository( const CStaticRepository &) {}
	CStaticRepository& operator=( const CStaticRepository&)
	{ return *this; }

};

// ==================== Server =====================

class CStaticServer
	: public CServer
{

public:

	CStaticServer( TInt aPriority);
	~CStaticServer()
	{}
	TAny* GetStatics();

	TInt Start( const TDesC& aName,
			const StaticCreateFunction aStatFun);

	void IncrementSessions()
	{ iSessionCount ++ ; }
	void DecrementSessions();

private:

	TInt	iSessionCount;
	TAny*	iIsStaticRepository;
	StaticCreateFunction	iStatFun;

#ifdef __PSISOFT32__	/* { ER5 */
	CSession* NewSessionL(RThread aClient,
			const TVersion& aVersion) const;
#else	/* }{ ER6 */
	CSharableSession* NewSessionL( const TVersion& aVersion) const;
#endif	/* }{ __PSISOFT32__ */

	CStaticServer( const CStaticServer&)
		: CServer( EPriorityStandard)
	{}
	CStaticServer& operator =( const CStaticServer&)
	{ return *this; }

};

class CStaticSession
	: public CSession
{

public:

	CStaticSession::~CStaticSession()
	{
		iServer->DecrementSessions();
	}

	virtual void ServiceL( const RMessage& m);
	static CStaticSession* NewL( RThread& aClient,
			CStaticServer* aServer);

private:

	CStaticServer* iServer;

	CStaticSession( RThread aClient, CStaticServer* aServer)
		: CSession( aClient), iServer( aServer)
	{
		iServer->IncrementSessions();
	}

	CStaticSession( const CStaticSession& aCs)
		: CSession( aCs.Client())
	{}
	CStaticSession& operator =( const CStaticSession&)
	{ return *this; }

	// in order to instantiate the class
public:
	CStaticSession() : CSession( RThread()), iServer(0) {}
	void dummy_func() {}
friend TAny* GetStaticsRepository( const StaticCreateFunction aStatFun,
		const TDesC& aServerName);
};

// ================ Client ==================

struct SDllTlsData
{
	TAny*	iStatics;
	TAny*	iDllTls;
	CStaticRepository*	iSession;
	SDllTlsData()
		: iStatics( 0),
		  iDllTls( 0),
		  iSession( 0)
	{}
	~SDllTlsData()
	{
		if( iSession)
		{
			iSession->Close();
			delete iSession;
		}
	}
};

TAny** GetDllTls()
{
	SDllTlsData* d = (SDllTlsData*)Dll::Tls();
	if( d == 0)
	{
		d = new SDllTlsData;
		if( !d)
			User::Panic( _L("Error allocating EpocStat Tls data"), 1);
		Dll::SetTls( d);
	}
	return( &(d->iDllTls));
}

TAny* GetStaticsRepository( StaticCreateFunction aStatFun,
		const TDesC& aServerName)
{
	SDllTlsData* d = (SDllTlsData*)Dll::Tls();
	if( d == 0)
	{
		d = new SDllTlsData;
		if( !d)
			User::Panic( _L("Error allocating EpocStat Tls data"), 2);
		Dll::SetTls( d);
	}
	if( d->iStatics == 0)
	{
		// Prepare server name - unique per DLL per process
		RProcess me;
		TProcessId process_id = me.Id();
#ifdef __PSISOFT32__	/* { ER5 */
		TUint pid = *((TUint*)&process_id);
#else	/* }{ __PSISOFT32__ */
		TUint pid = process_id;
#endif	/* } __PSISOFT32__ */
		TName srv_name;
		srv_name.Format( _L("%S_%u"), &aServerName, pid);
		// Connect to the server
		d->iSession = new CStaticRepository;
		TInt err = d->iSession->Connect( srv_name, aStatFun);
		if( err != KErrNone)
		{
			User::Panic( _L("Error connecting to EpocStat"), err);
			return( 0);	// to make compiler happy
		}
		d->iStatics = d->iSession->GetStatics();
		if( !d->iStatics)
			User::Panic( _L("Illegal zero EpocStat data"), 0);
	}
	return( d->iStatics);
}

TAny* CStaticRepository::GetStatics()
{
	// request to get static
	TInt p[KMaxMessageArguments]; // TInt p[4];
	TAny * staticVar = 0;
	TPtr8 tmp( (TUint8*)&staticVar, sizeof( staticVar));
	p[0] = (TInt)&tmp;
	SendReceive( EGetStatics, p);
	return( staticVar);
}

TInt CStaticRepository ::Connect( const TDesC& aName,
		const StaticCreateFunction aStatFun)
{
	const TVersion dummy;
	TInt err = CreateSession( aName, dummy, 4);
	if( ( err == KErrNotFound) || ( err == KErrServerTerminated))
	{
		// no such server, thus we have to start it
		err = StartServerThread( aName, aStatFun);
		if( ( err != KErrNone) && ( err != KErrAlreadyExists))
		{
			return( err);
		}
		// more trials to connect to server existing for sure...
		for( TInt i = 0; i < 3; i ++ )
		{
			err = CreateSession( aName, dummy, 4);
			if( err == KErrNone)
				break;
			User::After( 10000);
		}
	}
	return( err);
}

TInt CStaticRepository::ThreadFunction( TAny* anArg)
{
	SThreadInitData* init_data = (SThreadInitData*)anArg;
	RThread father;
	TInt err = father.Open( init_data->tid);
	if( err != KErrNone)
	{
		User::Panic( _L("Error opening father thread"), err);
	}

	// start scheduler and server
	CStaticServer* cs = 0;
	CActiveScheduler *pA = new CActiveScheduler;
	if( pA == 0)
	{
		father.RequestComplete( init_data->status,
				KErrNoMemory);
	}
	else
	{
		CActiveScheduler::Install(pA);
		cs = new CStaticServer( CActive::EPriorityStandard);
		if( cs == 0)
		{
			father.RequestComplete( init_data->status,
					KErrNoMemory);
		}
		else
		{
			err = cs->Start( init_data->server_name,
					init_data->statFun);
			father.RequestComplete( init_data->status, err);
			if( err == KErrNone)
			{
				CActiveScheduler::Start();
			}
		}
	}
	father.Close();
	if( cs)
		delete cs;
	if( pA)
		delete pA;
	return( err);
}

TInt CStaticRepository::StartServerThread( const TDesC& aName,
					const StaticCreateFunction aStatFun)
{
	// Symbian OS should has threads with unique full names.
	// Unfortunately Serier 60 Emulator (at least, perhaps others
	// too) allows several threads with the same full name. So it's
	// possible that we will create several threads, but all
	// "secondary" will attempt to create server, it will exist
	// aready (from the 1st such thread), and will terminate.
	RThread thread;
	TRequestStatus status = KRequestPending;
	SThreadInitData init_data( aName, aStatFun);
	init_data.tid = thread.Id();
	init_data.status = &status;
	TInt res = thread.Create( aName,   // create new server thread
		ThreadFunction, // thread's main function
		KDefaultStackSize,
		KMinHeapSize,
		0x10000,
		&init_data);// passed as argument to thread function
	if (res == KErrNone) // thread created ok - now start it going
	{
		thread.SetPriority( EPriorityNormal);
		thread.Resume(); // start it going
		User::WaitForRequest( status);
		res = status.Int();
	}
	thread.Close();

	// misterious call for emulator:
	// notify the kernel that a server has started.
#if defined (__WINS__)
	if( res == KErrNone)
		UserSvr::ServerStarted();
#endif

	return res;
}


// ==================== Server =====================

// ==================== CStaticServer ==============

CStaticServer::CStaticServer( TInt aPriority)
	: CServer( aPriority),
	  iSessionCount( 0),
	  iIsStaticRepository( 0),
	  iStatFun( 0)

{
	// empty
}

#ifdef __PSISOFT32__	/* { ER5 */

CSession* CStaticServer::NewSessionL( RThread aClient, const TVersion& aVersion) const
{
	// as we are working in the very close client/server, we ignore version
	return CStaticSession::NewL(aClient, (CStaticServer*)this);
}

#else	/* }{ ER6 */

CSharableSession* CStaticServer::NewSessionL( const TVersion&) const
{
	// as we are working in the very close client/server, we ignore version
	RThread aClient = Message().Client();
	return CStaticSession::NewL(aClient, (CStaticServer*)this);
}

#endif	/* }{ __PSISOFT32__ */


TAny* CStaticServer::GetStatics()
{
	if( iIsStaticRepository == NULL)
		iIsStaticRepository = iStatFun();
	return iIsStaticRepository;
}

TInt CStaticServer::Start( const TDesC& aName, const StaticCreateFunction aStatFun)
{
	iStatFun = aStatFun;
#ifdef __PSISOFT32__	/* { ER5 */
	iName = aName.Alloc();
	return CServer::Start();
#else	/* }{ ER6 */
	return CServer::Start( aName);
#endif	/* }{ __PSISOFT32__ */
}

void CStaticServer::DecrementSessions()
{
	iSessionCount -- ;
	if( iSessionCount <= 0)
		CActiveScheduler::Stop();
}

//======================= CStaticSession ====================

void CStaticSession::ServiceL( const RMessage& m)
{
	switch( m.Function())
	{
	// These 2 cases:
	// case EConnect
	// case EDisConnect
	// are handled by base class
	case EGetStatics:
	{
		CStaticServer* s = (CStaticServer*)Server();
		TAny* statVar = s->GetStatics();
		TPtrC8 tmp( (const TUint8*)&statVar, sizeof( statVar));
		Client().WriteL( m.Ptr0(), tmp, 0);
		m.Complete( KErrNone);
	}
	default:
		break;
	}
}

CStaticSession* CStaticSession::NewL( RThread &aClient, CStaticServer * aServer)
{
	return new (ELeave) CStaticSession( aClient, aServer);
}

// ====== TLS handling ===========
//
// Regarding ER5:
//
// DLL entry point function.
//
// The behaviour on the WINS platform is different to that on the
// ARM platform.
//
// If the DLL is attaching to a new thread, then a new SDllTlsData
// object is constructed; thread local storage is set to point to
// this SDllTlsData object.
//
// Thread local storage is reset when the DLL detaches from the thread.
//
// If the DLL is attaching to a new process, then a new SDllTlsData
// object is constructed; thread local storage is set to point to
// this SDllTlsData object.
//
// Under WINS thread local storage cannot be reset when the process
// detaches from the DLL.
//
// Notes:
// 1. Here, E32Dll is called with EDllThreadAttach after it is called
//    with EDllProcessAttach. Thread local storage for the primary
//    thread should only be set when handling EDllProcessAttach
// 2. E32Dll is never called with EDllThreadAttach
//
// It seems that starting Symbian OS v.6 this behavior is fixed.

GLDEF_C TInt E32Dll( TDllReason aReason)
{
	SDllTlsData* d = (SDllTlsData*)Dll::Tls();
	switch( aReason)
	{
	case EDllProcessAttach:
	case EDllThreadAttach:
		if( !d)
		{
			d = new SDllTlsData;
			if( !d)
				User::Panic( _L("Error allocating EpocStat Tls data"), 0);
			Dll::SetTls( d);
		}
		break;
	case EDllThreadDetach:
	case EDllProcessDetach:
		if( d)
			delete d;
		Dll::SetTls( 0);
		User::After( 250000);	// 0.25 sec - to give server a chance to shutdown
		break;
	}
	return( KErrNone);
}

