/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Image.c,v 1.12 2005/08/21 20:09:38 fdie Exp $ */

#include "superwaba.h"
#include "SDL.h"
#include "SW_natives.h"
#include "SW_logging.h"

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#if HAVE_MATH_H
#include <math.h>
#endif

/* from SW_Graphics.c */
extern GrHandle activeGrHandle;
extern GrHandle mainGrHandle;
extern WClass * imageClass;

GrHandle makeGrHandle(SDL_Surface * surface, Sint32 videoFlag);
void setDrawWindow(GrHandle newGrHandle);
void changeColors(SDL_Surface * surface,
                  Sint32 x, Sint32 y, Sint32 width, Sint32 height,
                  Pixel pixelFrom, Pixel pixelTo);

/*--------------------------------------------------------------createSurface-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Creates a new GrHandle for a bitmap (or screen) with the given            |
|   width, height.                                                            |
|   - if hParent is 1, a new offscreen bitmap is allocated, compatible        |
|   with the format of the video                                              |
|   - if hParent is a graphic handle, a new offscreen bitmap is allocated,    |
|   compatible with the format of the hParent graphics                        |
+----------------------------------------------------------------------------*/
static GrHandle createSurface(WObject obj, Sint32 width, Sint32 height, GrHandle hParent)
{
   GrHandle grhOut;
   Sint32 videoFlag;
   SDL_Surface * target;
   SDL_Surface * parent;
   SDL_PixelFormat * parentFormat;

   _LOG_INFO((LOG_CAT_IMG, "createSurface(%ld,%ld)", width, height));
   if (hParent == (GrHandle)1)
   {
      parent = SDL_GetVideoSurface();
      videoFlag = VIDEOFLAG_COMPATIBLE_OFFSCREEN;
   }
   else
   {
      parent = ((SDL_Surface *)(GRHANDLE_WinHandle(hParent)));
      videoFlag = GRHANDLE_VideoFlag(hParent);
   }
   parentFormat = parent->format;
   if (!obj || (*vmGlobals->pushObject)(obj) == -1)
      return 0;        // GC can be called on createSurface
   target = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height, parentFormat->BitsPerPixel,
                                 parentFormat->Rmask, parentFormat->Gmask, parentFormat->Bmask, parentFormat->Amask);
   if (!target)
      return 0;
   if (parentFormat->palette && target->format->palette)
   {
      memcpy(target->format->palette->colors, parentFormat->palette->colors, parentFormat->palette->ncolors * sizeof(SDL_Color));
      target->format->palette->ncolors = parentFormat->palette->ncolors;
   }
   (*vmGlobals->popObject)(obj);
   grhOut = makeGrHandle(target, videoFlag);
   if (grhOut == 0)
      SDL_FreeSurface(target);
   return grhOut;
}

/*---------------------------------------------------------createSimilarImage-+
|                                                                             |
+----------------------------------------------------------------------------*/
static WObject createSimilarImage(WObject imageSrc, int width, int height)
{
   GrHandle srcG = WOBJ_ControlImageGraphicsGrHandle(imageSrc);
   GrHandle dstG;
   WObject imageDst;

   _LOG_INFO((LOG_CAT_IMG, "createSimilarImage(%d,%d)", width, height));
   // cache class pointers for performance
   if (!imageClass) imageClass = vmGlobals->getClass(vmGlobals->createUtfString("waba/fx/Image"),1);
   // create the new imageDst object with the given sizes
   imageDst = (*vmGlobals->createObject)(imageClass);
   if (!imageDst)
      return 0;
   dstG = createSurface(imageDst, width, height, srcG);
   if (!dstG)
      return 0; // no more memory?
   WOBJ_ImageWidth(imageDst) = width;
   WOBJ_ImageHeight(imageDst) = height;
   WOBJ_ControlImageGraphicsGrHandle(imageDst) = dstG;
   WOBJ_ImageTranspPixel(imageDst) = WOBJ_ImageTranspPixel(imageSrc); // guich@410_8: forgot to copy those attributes
   WOBJ_ImagePalette(imageDst) = WOBJ_ImagePalette(imageSrc);
   return imageDst;
}

/*--------------------------------------------------------------------drawRow-+
|                                                                             |
+----------------------------------------------------------------------------*/
void drawRow(Sint32 xStart, Sint32 xEnd, Sint32 y,
             Uint8 *rgbOrIndex, Uint32 *colorMap)
{
   Pixel * tgt;
   SDL_Surface * targetSurface = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
   _LOG_DEBUG((LOG_CAT_IMG, "drawRow(%ld,%ld - %ld)", xStart, xEnd, y));

   SDL_LockSurface(targetSurface);
   tgt = (Pixel *)((char *)targetSurface->pixels + (y * targetSurface->pitch)) + xStart;
   if (colorMap == 0)
   {
      // colorMap is null if the device has more than 256 colors
      // and rgbOrIndex is an array of 24 bits RGB.
      SDL_PixelFormat const * format = targetSurface->format;
      Sint32 x;
      for (x = xStart; x < xEnd; ++x)
      {
#if 0
         *tgt++ = (((rgbOrIndex[0] >> format->Rloss) << format->Rshift) |
                   ((rgbOrIndex[1] >> format->Gloss) << format->Gshift) |
                   ((rgbOrIndex[2] >> format->Bloss) << format->Bshift) |
                   format->Amask);
#else
         *tgt++ = SDL_MapRGB(SDL_GetVideoSurface()->format, 
                                (Uint8)rgbOrIndex[0], 
                                (Uint8)rgbOrIndex[1], 
                                (Uint8)rgbOrIndex[2]);
#endif                   
         rgbOrIndex += 3;
      }
   }
   else
   {
      // otherwise, "colormap" is the Palette of the target surface
      // then rgbOrIndex is an index into the colormap.
      memcpy(tgt, rgbOrIndex, xEnd - xStart);
   }
   SDL_UnlockSurface(targetSurface);
}

/*---------------------------------------------------------graphicsStartImage-+
|                                                                             |
+----------------------------------------------------------------------------*/
void graphicsStartImage(Uint32 *ret, Uint32 *args)  /* PUBLIC_GraphicsStartImage */
{
   // NOTE: it is assumed that the image is brand new, and was created using a 'NULL' constructor, so we don't have to free an existing bitmap.
   WObject image = (WObject)args[0];
   Sint32 width = WOBJ_ImageWidth(image);
   Sint32 height = WOBJ_ImageHeight(image);

   _LOG_INFO((LOG_CAT_IMG, "graphicsStartImage(%ld,%ld)", width, height));
   *ret = 0;
   if ((width > 0) && (height > 0))
   {
      GrHandle oldG, newG;
      DrawRowProcType * pDrawRow = (DrawRowProcType *)args[2];

      newG = createSurface(image, width, height, (GrHandle)1);
      if (!newG) return; // no more memory?
      WOBJ_ControlImageGraphicsGrHandle(image) = newG;
      oldG = activeGrHandle;
      setDrawWindow(newG);
      *pDrawRow = drawRow;
      *(GrHandle*)ret = oldG;
#ifndef WINCE // this does not work well on wince
      if (args[1] != 0)     // redoTransparentPixel
      {
         /*
         | redoTransparentPixel means that
         | 1) a transparent Pixel has been defined for this image; and,
         | 2) its value is a RGB 888 value (not an index in the colormap).
         */
         SDL_PixelFormat *format = ((SDL_Surface *)GRHANDLE_WinHandle(newG))->format;
         Uint32 tp, tpC = WOBJ_ImageTranspPixel(image);
         _LOG_DEBUG((LOG_CAT_IMG, "redoTransparentPixel = (%02x,%02x,%02x)", 
                                    (int)((tpC >> 16) & 0xFF), 
                                    (int)((tpC >>  8) & 0xFF), 
                                    (int)( tpC        & 0xFF)));

//         tp = ((((tp >>16) & 0xFF) >> format->Rloss) << format->Rshift) |
//              ((((tp >> 8) & 0xFF) >> format->Gloss) << format->Gshift) |
//              (((tp        & 0xFF) >> format->Bloss) << format->Bshift) |
//              format->Amask;

         tp = SDL_MapRGB(format, 
                             (Uint8)((tpC >> 16) & 0xFF), 
                             (Uint8)((tpC >>  8) & 0xFF), 
                             (Uint8)( tpC        & 0xFF));
              
         _LOG_DEBUG((LOG_CAT_IMG, "pixel=>(0x%08lx)", tp));
         WOBJ_ImageTranspPixel(image) = tp;
      }
#endif
   }
}

/*-----------------------------------------------------------graphicsEndImage-+
|                                                                             |
+----------------------------------------------------------------------------*/
void graphicsEndImage(GrHandle oldG)    /* PUBLIC_GraphicsEndImage */
{
   _LOG_INFO((LOG_CAT_IMG, "graphicsEndImage"));
   if (oldG) setDrawWindow(oldG);
}

/*--------------------------------------------------------------------loadBMP-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void loadBMP(WObject image, Uint8 * src, int size)
{
   SDL_Surface * bmpSurface;
   _LOG_INFO((LOG_CAT_IMG, "loadBMP size=%d", size));
#if defined(SUPPORT_PUBLIC_SDL) // fdie@ SDL still doesn't support compressed BMP files
   SDL_Surface *SW_LoadBMP_RW(SDL_RWops *src, int freesrc);
   bmpSurface = SW_LoadBMP_RW(SDL_RWFromMem(src, size), 1);
#else   
   bmpSurface = SDL_LoadBMP_RW(SDL_RWFromMem(src, size), 1); // use our own patched SDL with rle support
#endif   
   if (bmpSurface)
   {
      /* Set up the fields in the instance of the Java Image */
      //>>>PGR: TESTME set VIDEOFLAG_DIFFERENT_OFFSCREEN and to avoid this lengthy (?) conversion
      SDL_Surface * imgSurface = SDL_DisplayFormat(bmpSurface);
      GrHandle imgHandle = makeGrHandle(imgSurface, VIDEOFLAG_COMPATIBLE_OFFSCREEN);

      WOBJ_ControlImageGraphicsGrHandle(image) = imgHandle;
      WOBJ_ImageWidth(image) = imgSurface->w;
      WOBJ_ImageHeight(image) = imgSurface->h;
      SDL_FreeSurface(bmpSurface);
   }
}

/*-----------------------------------------------------------------scaleImage-+
|                                                                             |
| Impl Note (PGR)                                                             |
|                                                                             |
| This is a simple scaling algorithm.                                         |
|                                                                             |
| We first compute a "Pixel Diffusion" table:                                 |
|    kx          ky                                                           |
|    |           |                                                            |
|    V           V                                                            |
|    [0, 7, 4], [3, 5]                                                        |
|                                                                             |
| means that for the 3 x 2 source pixels                                      |
| reading the source row 0, do 3 times on the 3 first target rows             |
|    skip the source pixel at col 0                                           |
|    do 7 times on the 7 first cols: target pixel = source pixel at col 1     |
|    do 4 times on the 4 next cols: target pixel = source pixel at col 2      |
| reading the source row 1, do 5 times on the 5 next target rows              |
|    skip the source pixel at col 0                                           |
|    do 7 times on the 7 first cols: target pixel = source pixel at col 1     |
|    do 4 times on the 4 next cols: target pixel = source pixel at col 2      |
|                                                                             |
| ABC   becomes    BBBCCCC                                                    |
| DEF              BBBCCCC                                                    |
|                  BBBCCCC                                                    |
|                  EEEFFFF                                                    |
|                  EEEFFFF                                                    |
|                  EEEFFFF                                                    |
|                  EEEFFFF                                                    |
|                  EEEFFFF                                                    |
|                                                                             |
+----------------------------------------------------------------------------*/
Uint8 scaleImage(SDL_Surface * sourceSurface, Uint16 width, Uint16 height,
                 SDL_Surface * targetSurface, Uint16 newWidth, Uint16 newHeight)
{
   Sint32 sourcePitch = sourceSurface->pitch;
   Sint32 targetPitch = targetSurface->pitch;
   Sint32 * buf = (Sint32 *)(*vmGlobals->xmalloc)(sizeof(Sint32) * (width + height) << 1);
   Sint32 x, y;
   Sint32 kx = 0;
   Sint32 ky = width;
   Sint32 kw = ky + height;
   Sint32 kh = kw + width;
   Sint32 bytespp = sourceSurface->format->BytesPerPixel;
   Uint8 * sourcePixels;

   _LOG_INFO((LOG_CAT_IMG, "scaleImage %dx%d => %dx%d", width, height, newWidth, newHeight));
   if (!buf) return 0; // if no memory to allocate the buffer, we simply return.

   // create a temporary "Pixel Diffusion" table:
   buf[kx] = 0;
   for (x=1; ; ++x)
   {
      Sint32 temp = (x * newWidth) / width;
      buf[kw + x-1] = temp - buf[kx + x-1];
      if (x == width) break;
      buf[kx + x] = temp;
   }
   buf[ky] = 0;
   for (y=1; ; ++y)
   {
      Sint32 temp = (y * newHeight) / height;
      buf[kh + y-1] = temp - buf[ky + y-1];
      if (y == height) break;
      buf[ky + y] = temp;
   }

   // get and set the pixels according to the diffusion table
   // Note: the surface are deemed to be compatible!
   SDL_LockSurface(sourceSurface);
   SDL_LockSurface(targetSurface);

   for (sourcePixels = sourceSurface->pixels, y=0; y < height; ++y, sourcePixels = sourcePixels + sourcePitch)
   {
      Sint32 h0 = buf[kh + y];
      Sint32 yy0 = (buf[ky + y] * targetPitch);
      Uint8 * pSource = sourcePixels;
      for (x=0; x < width; ++x, pSource += bytespp)
      {
         Sint32 w0 = buf[kw + x];
         Sint32 h = h0;
         Uint8 * targetPixels;
         for (targetPixels = (char *)targetSurface->pixels + yy0; h-- > 0; targetPixels = (char *)targetPixels + targetPitch)
         {
            Sint32 w = w0;
            Uint8 * pTarget = targetPixels + (buf[kx+x] * bytespp);
            switch (bytespp)
            {
               case 1:
                  while (w-- > 0)
                     *pTarget++ = pSource[0];
                  break;
               case 2:
                  while (w-- > 0)
                  {
                     *((Uint16 *)pTarget)++ = *(Uint16 *)pSource;
                  }
                  break;
               case 3:
                  while (w-- > 0)
                  {
                     *pTarget++ = pSource[0];
                     *pTarget++ = pSource[1];
                     *pTarget++ = pSource[2];
                  }
                  break;
               case 4:
                  while (w-- > 0)
                     *((Uint32 *)pTarget)++ = *(Uint32 *)pSource;
                  break;
            }
         }
      }
   }

   SDL_UnlockSurface(sourceSurface);
   SDL_UnlockSurface(targetSurface);
   // updateRectangle(targetSurface, 0, 0, newWidth, newHeight);
   xfree(buf);
   return 1;  // success
}

/*---------------------------------------------------------------imageDestroy-+
|                                                                             |
+----------------------------------------------------------------------------*/
void imageDestroy(WObject image)
{
   GrHandle h = WOBJ_ControlImageGraphicsGrHandle(image);

   if (h)
   {
      if (h == activeGrHandle)  // guich@200b3: is this the active GrHandle?
         setDrawWindow(mainGrHandle); // switch to other handle
      SDL_FreeSurface((SDL_Surface *)(GRHANDLE_WinHandle(h)));
      if (GRHANDLE_Palette(h)) xfree(GRHANDLE_Palette(h));
      (*vmGlobals->xfree_)(h);
      WOBJ_ControlImageGraphicsGrHandle(image) = NULL;
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void imageFree(Image obj)
void ImageFree(Var stack[]) // (Lwaba/fx/Image;)V
{
   WObject image = stack[0].obj;
   _LOG_INFO((LOG_CAT_IMG, "ImageFree"));
   imageDestroy(image);
   WOBJ_ImageWidth(image) = 0;
   WOBJ_ImageHeight(image) = 0;
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void imageLoad(Image obj, String path)
void ImageLoad(Var stack[]) // (Lwaba/fx/Image;Ljava/lang/String;)V
{
   WObject image = stack[0].obj;
   UtfString path = (*vmGlobals->stringToUtf)(stack[1].obj, STU_USE_STATIC);
   Uint16 pathLen = (Uint16)path.len;

   _LOG_INFO((LOG_CAT_IMG, "ImageLoad(%s)", path.str));

   /*
   | NOTE: we don't have to free an existing bitmap because
   | this is only called from an image constructor
   */
   WOBJ_ControlImageGraphicsGrHandle(image) = NULL;
   WOBJ_ImageWidth(image) = 0;
   WOBJ_ImageHeight(image) = 0;

   if (path.len)
   {
      uint32 size;
      uchar * p = (*vmGlobals->lockWarpRec)(path.str, (Uint16)path.len, &size);
      if (p)
      {
         loadBMP(image, p, size);
         (*vmGlobals->unlockWarpRec)(p, pathLen);
      }
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void imageParse(Image obj, byte []fullBmpDescription)
void ImageParse(Var stack[]) // (Lwaba/fx/Image;[B)V
{
   WObject image = stack[0].obj; // class object
   WObject bytes = stack[1].obj;

   _LOG_INFO((LOG_CAT_IMG, "ImageParse"));
   WOBJ_ControlImageGraphicsGrHandle(image) = NULL;
   WOBJ_ImageWidth(image) = 0;
   WOBJ_ImageHeight(image) = 0;
   loadBMP(image, (Uint8 *)WOBJ_arrayStart(bytes), WOBJ_arrayLen(bytes));
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void imageCreate(Image obj)
void ImageCreate(Var stack[]) // (Lwaba/fx/Image;)V
{
   WObject image;
   Sint16 width, height;
   GrHandle imgHandle;

   image = stack[0].obj;
   width = (Sint16)WOBJ_ImageWidth(image);
   height = (Sint16)WOBJ_ImageHeight(image);
   _LOG_INFO((LOG_CAT_IMG, "ImageCreate(%d,%d)", width, height));

   if (width > 0 && height > 0)
      imgHandle = createSurface(image, width, height, (GrHandle)1);
   else
      imgHandle = NULL;
   WOBJ_ControlImageGraphicsGrHandle(image) = imgHandle;
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void imageChangeColors(Image obj, int fromR, int fromG, int fromB, int toR, int toG, int toB)
void ImageChangeColors(Var stack[]) // (Lwaba/fx/Image;IIIIII)V
{
   WObject image = stack[0].obj; // class object
   Sint32 width = (Sint32)WOBJ_ImageWidth(image);
   Sint32 height = (Sint32)WOBJ_ImageHeight(image);
   GrHandle imgHandle = WOBJ_ControlImageGraphicsGrHandle(image);

   _LOG_INFO((LOG_CAT_IMG, "ImageChangeColors"));
   if (imgHandle)
   {
      SDL_Surface * surf = (SDL_Surface *)GRHANDLE_WinHandle(imgHandle);
      SDL_PixelFormat * fmt = surf->format;
      changeColors(surf, 0, 0, width, height,
         (Pixel)SDL_MapRGB(fmt, (Uint8)stack[1].intValue, (Uint8)stack[2].intValue, (Uint8)stack[3].intValue), // pixelFrom
         (Pixel)SDL_MapRGB(fmt, (Uint8)stack[4].intValue, (Uint8)stack[5].intValue, (Uint8)stack[6].intValue)); // pixelTo
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods waba.fx.Image imageGetScaledInstance(Image obj, int newWidth, int newHeight)
void ImageGetScaledInstance(Var stack[]) // (Lwaba/fx/Image;II)Lwaba/fx/Image;
{
   WObject imageSrc = stack[0].obj;
   Sint16 newWidth = (Sint16)stack[1].intValue;
   Sint16 newHeight = (Sint16)stack[2].intValue;
   Sint16 width = (Sint16)WOBJ_ImageWidth(imageSrc);
   Sint16 height = (Sint16)WOBJ_ImageHeight(imageSrc);
   GrHandle srcG = WOBJ_ControlImageGraphicsGrHandle(imageSrc);

   _LOG_INFO((LOG_CAT_IMG, "ImageGetScaledInstance"));
   if (srcG && (width > 0) && (height > 0))
   {
      WObject imageDst = createSimilarImage(imageSrc, newWidth, newHeight);
      if (imageDst)
      {
         GrHandle dstG = WOBJ_ControlImageGraphicsGrHandle(imageDst);
         if (scaleImage((SDL_Surface *)GRHANDLE_WinHandle(srcG), width, height,
                        (SDL_Surface *)GRHANDLE_WinHandle(dstG), newWidth, newHeight))
         {
            RETURN32(imageDst);
         }
         imageDestroy(imageDst);
      }
   }
   RETURN32(0);
}


//////////////////////////////////////////////////////////////////////////
// this code must be better written: remove the dependancy with the SW_graphics's 
// hashtable code and support other pixel formats.
struct htElemColor * htPut(struct htElemColor * ht[], Uint32 key, Sint32 value);
struct htElemColor * htGet(struct htElemColor * ht[], Uint32 key);
void disposeHashtable(struct htElemColor *ht[], Uint8 deleteObj);
SDL_Color* getStaticWebPalette();
struct htElemColor
{
   Uint32 key;
   Sint32 value;
   struct htElemColor * next;
};

// waba/fx/NativeMethods void imageGetPixelRow(Image obj, byte []fillIn, int y)
void ImageGetPixelRow(Var stack[]) // (Lwaba/fx/Image;[BI)V
{
   static WObject lastImg = 0;
   WObject image = stack[0].obj; // class object
   WObject fillIn;
   Sint32 y;
   Uint8 *dst;
   Uint8 *src;
   GrHandle h;
   SDL_Surface * surf;
   int32 dstsize;

   static struct htElemColor * ht[0xFF+1];
   static SDL_PixelFormat palformat;
   static SDL_Palette pal;
   struct htElemColor * elem;

   _LOG_INFO((LOG_CAT_IMG, "ImageGetPixelRow"));
   if (lastImg != image)
   {
      disposeHashtable(ht, 0); // changed images, reset the hashtable
      lastImg = image;
   }

   if (palformat.BytesPerPixel == 0) // prepare the structure
   {
      pal.ncolors = 256;
      pal.colors = getStaticWebPalette();
	   palformat.BitsPerPixel = 8;
	   palformat.BytesPerPixel=1;
	   palformat.Rloss=0;
	   palformat.Gloss=0;
	   palformat.Bloss=0;
	   palformat.Aloss=0;
	   palformat.Rshift=16;
	   palformat.Gshift=8;
	   palformat.Bshift=0;
	   palformat.Ashift=24;
	   palformat.Rmask=255;
	   palformat.Gmask=255;
	   palformat.Bmask=255;
	   palformat.Amask=255;
	   palformat.colorkey=0;
	   palformat.alpha=0;
      palformat.palette = &pal;
   }       

   fillIn = stack[1].obj;
   y = stack[2].intValue;
   dst = (Uint8 *)WOBJ_arrayStart(fillIn);
   dstsize = WOBJ_arrayLen(fillIn);
   h = WOBJ_ControlImageGraphicsGrHandle(image);

   surf = (SDL_Surface *)GRHANDLE_WinHandle(h);
   SDL_LockSurface(surf);

   src = (Uint8*)((char *)surf->pixels + (y * surf->pitch));
   {
   SDL_PixelFormat *fmt = surf->format;
   Uint8 rloss  = fmt->Rloss;
   Uint8 gloss  = fmt->Gloss;
   Uint8 bloss  = fmt->Bloss;
   Uint8 rshift = fmt->Rshift;
   Uint8 gshift = fmt->Gshift;
   Uint8 bshift = fmt->Bshift;
   Uint32 rmask = fmt->Rmask;
   Uint32 gmask = fmt->Gmask;
   Uint32 bmask = fmt->Bmask;
   Sint32 bytespp = fmt->BytesPerPixel;
   for (; dstsize-- > 0; dst++)
   {
      int32 key;
      switch (bytespp)
      {
         case 1: // already palettized
            *dst++ = *src++;
            continue;
         case 2:
            key = READ_LE_UINT16(src);
            src += 2;
            break;
         case 3:
            if (SDL_BYTEORDER == SDL_BIG_ENDIAN) // a good compiler will strip the not used code block
               key = ((Sint32)src[0] << 16) | ((Sint32)src[1] <<  8) | src[2];
            else
               key = ((Sint32)src[2] << 16) | ((Sint32)src[1] <<  8) | src[0];
            src += 3;
            break;
         case 4:
            key = READ_LE_UINT32(src);
            src += 4;
            break;
      }

      elem = htGet(ht, key);
      if (elem)
         *dst = (Uint8)(elem->value);
      else
      {
         Sint32 r = ((key & rmask) >> rshift) << rloss;
         Sint32 g = ((key & gmask) >> gshift) << gloss;
         Sint32 b = ((key & bmask) >> bshift) << bloss;
         Pixel val = SDL_MapRGB(&palformat, (Uint8)r, (Uint8)g, (Uint8)b);
         htPut(ht, key, val);
         *dst = (Uint8)val;
      }
   }
   }
   SDL_UnlockSurface(surf);
}

/*-------------------------------------------------------setCompatibleSurface-+
|                                                                             |
+----------------------------------------------------------------------------*/
static SDL_Surface * setCompatibleSurface(WObject img)
{
   GrHandle grh = WOBJ_ControlImageGraphicsGrHandle(img);
   _LOG_INFO((LOG_CAT_IMG, "setCompatibleSurface"));
   if (GRHANDLE_VideoFlag(grh) == VIDEOFLAG_DIFFERENT_OFFSCREEN)
   {
      SDL_Surface * surfIn = (SDL_Surface *)GRHANDLE_WinHandle(grh);
      SDL_Surface * surfOut = SDL_DisplayFormatAlpha(surfIn);
      GRHANDLE_WidthInBPP(grh) = surfOut->pitch;
      GRHANDLE_VideoFlag(grh) = VIDEOFLAG_COMPATIBLE_OFFSCREEN;
      GRHANDLE_WinHandle(grh) = surfOut;
      SDL_FreeSurface(surfIn);
      return surfOut;
   }
   return (SDL_Surface *)GRHANDLE_WinHandle(grh);
}

/*----------------------------------------------ImageGetRotatedScaledInstance-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/fx/NativeMethods waba.fx.Image imageGetRotatedScaledInstance(waba.fx.Image obj, int percScale, int angle, waba.fx.Color fill)
void ImageGetRotatedScaledInstance(Var stack[]) // (Lwaba/fx/Image;IILwaba/fx/Color;)Lwaba/fx/Image;
{
   WObject imageIn = stack[0].obj;
   Sint32 perScale = stack[1].intValue;
   Sint32 angle = stack[2].intValue;
   WObject color = stack[3].obj;

   SDL_Surface * surfIn;
   Pixel * pixelsIn;
   Sint32 pitchIn;
   Sint32 hIn;
   Sint32 wIn;
   Uint32 backColor;
   
   Sint32 rawSine;
   Sint32 rawCosine;
   Sint32 sine;
   Sint32 cosine;

   WObject imageOut;
   SDL_Surface * surfOut;
   Sint32 pitchOut;
   Pixel * pixelsOut;
   Sint32 hOut, wOut;
   Sint32 x0, y0;
   Sint32 i;

   _LOG_INFO((LOG_CAT_IMG, "ImageGetRotatedScaledInstance"));

   surfIn = setCompatibleSurface(imageIn);
   pixelsIn = (Pixel *)surfIn->pixels;
   pitchIn = surfIn->pitch;
   hIn = surfIn->h;
   wIn = surfIn->w;

   // if (perScale <= 0) return; // defense! (done on the java side)

   /* xplying by 0x10000 allow integer math, while not loosing much prec. */

   backColor = ((!color) ? WOBJ_ImageTranspPixel(imageIn)
                                : SDL_MapRGB(surfIn->format, (Uint8)WOBJ_ColorRedP(color),
                                             (Uint8)WOBJ_ColorGreenP(color), (Uint8)WOBJ_ColorBlueP(color)));
   angle = angle % 360;
   if ((angle % 90) == 0)
   {
      if (angle < 0) angle += 360;
      switch (angle)
      {
         case 0:
            rawSine = sine = 0;
            rawCosine = 0x10000;
            cosine = 0x640000 / perScale;
            break;
         case 90:
            rawSine = 0x10000;
            rawCosine = cosine = 0;
            sine = 0x640000 / perScale;
            break;
         case 180:
            rawSine = sine = 0;
            rawCosine = -0x10000;
            cosine = -0x640000 / perScale;
            break;
         default: // case 270:
            rawSine = -0x10000;
            rawCosine = cosine = 0;
            sine = -0x640000 / perScale;
            break;
      }
   }
   else
   {
      double rad = angle * 0.0174532925;
      rawSine = (Sint32)(sin(rad) * 0x10000);
      rawCosine = (Sint32)(cos(rad) * 0x10000);
      sine = (rawSine * 100) / perScale;
      cosine = (rawCosine * 100) / perScale;
   }

   /* create imageOut */
   {
      GrHandle grhOut;
      Sint32 x[3], y[3];
      Sint32 xMin = 0;
      Sint32 yMin = 0;
      Sint32 xMax = 0;
      Sint32 yMax = 0;
      x[0] = (wIn * rawCosine) >> 16;
      y[0] = (wIn * rawSine) >> 16;
      x[2] = (-hIn * rawSine) >> 16;
      y[2] = (hIn * rawCosine) >> 16;
      x[1] = x[0] + x[2];
      y[1] = y[0] + y[2];

      i = sizeof x / sizeof x[0];
      while (--i >= 0)
      {
         if (x[i] < xMin)
            xMin = x[i];
         else
         if (x[i] > xMax)
            xMax = x[i];

         if (y[i] < yMin)
            yMin = y[i];
         else
         if (y[i] > yMax)
            yMax = y[i];
      }
      wOut = ((xMax - xMin) * perScale) / 100;
      hOut = ((yMax - yMin) * perScale) / 100;
      imageOut = createSimilarImage(imageIn, wOut, hOut);
      if (!imageOut)
         RETURN32(0);
      /* center */
      x0 = ((wIn<<16) - (((xMax-xMin) * rawCosine) - ((yMax-yMin) * rawSine)) - 1) / 2;
      y0 = ((hIn<<16) - (((xMax-xMin) * rawSine) + ((yMax-yMin) * rawCosine)) - 1) / 2;

      grhOut = WOBJ_ControlImageGraphicsGrHandle(imageOut);
      surfOut = (SDL_Surface *)GRHANDLE_WinHandle(grhOut);
      pixelsOut = (Pixel *)surfOut->pixels;
      pitchOut = surfOut->pitch;
   }

   /* and draw! */
   SDL_LockSurface(surfIn);
   SDL_LockSurface(surfOut);
   while (--hOut >= 0)
   {
      Sint32 x = x0;
      Sint32 y = y0;
      Pixel * out = pixelsOut;
      for (i=wOut; --i >= 0; x += cosine, y += sine)
      {
         Sint32 u = x>>16;
         Sint32 v = y>>16;
         if (0 <= u && u < wIn && 0 <= v && v < hIn)
            *out++ = *(Pixel *)(((char *)pixelsIn + ((v * pitchIn) + (u * sizeof(Pixel)))));
         else
            *out++ = backColor;
      }
      x0 -= sine;
      y0 += cosine;
      pixelsOut = (Pixel *)((char *)pixelsOut + pitchOut);
   }
   SDL_UnlockSurface(surfIn);
   SDL_UnlockSurface(surfOut);
   RETURN32(imageOut);  // success
}

/*-------------------------------------------------------computeContrastTable-+
| Generates 128 points from a kind of symmetrical                             |
| yn = sum{1 to inf}(e-xt/tn)/dt,                                             |
| n being proportional to 'level'.                                            |
| The contrast level must be in the range -127 ... + 127 excluded             |
+----------------------------------------------------------------------------*/
static void computeContrastTable(Uint8 * table, Sint8 level)
{
   Sint32 i = 0;
   double factor;
   Uint8 * tbend = table+256;

   _LOG_INFO((LOG_CAT_IMG, "computeContrastTable"));
   if (level < 0)
   {
      level += 127;
      if (level <= 0) level = 1;  // defense!
      factor = (double)level / 127.0;
   }
   else
   {
      level = 127-level;
      if (level <= 0) level = 1;  // defense!
      factor = 127.0 / (double)level;
   }
   do
   {
      *--tbend = 255 - (*table++ = (Sint32)(127.0 * pow((double)i/127.0, factor)) & 0xff);
   } while (++i < 128);
}

/*--------------------------------------------------ImageGetTouchedUpInstance-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/fx/NativeMethods waba.fx.Image imageGetTouchedUpInstance(waba.fx.Image obj, byte brightness, byte contrast)
void ImageGetTouchedUpInstance(Var stack[]) // (Lwaba/fx/Image;BB)Lwaba/fx/Image;
{
   #define NO_TOUCHUP         0
   #define BRITE_TOUCHUP      1
   #define CTRST_TOUCHUP      2
   #define CTRSTBRITE_TOUCHUP 3  // OR'ed

   WObject imageIn = stack[0].obj;
   Sint8 iBrightness = stack[1].intValue;
   Sint8 iContrast = stack[2].intValue;
   Sint8 touchup = NO_TOUCHUP;
   SDL_Surface * surfIn;

   Pixel * pixelsIn;
   SDL_PixelFormat * fmt;
   Sint32 pitch;
   Sint32 w, h;

   WObject imageOut;
   GrHandle grhOut;
   SDL_Surface * surfOut;
   Pixel * pixelsOut;
   Uint8 table[256];
   Sint32 m, k, max;

   _LOG_INFO((LOG_CAT_IMG, "ImageGetTouchedUpInstance"));

   surfIn = setCompatibleSurface(imageIn);
   pixelsIn = (Pixel *)surfIn->pixels;
   fmt = surfIn->format;
   pitch = surfIn->pitch;
   w = surfIn->w;
   h = surfIn->h;

   imageOut = createSimilarImage(imageIn, w, h);

   if (!imageOut)
      RETURN32(0);
   grhOut = WOBJ_ControlImageGraphicsGrHandle(imageOut);
   surfOut = (SDL_Surface *)GRHANDLE_WinHandle(grhOut);
   pixelsOut = surfOut->pixels;

   if (iContrast)
   {
      touchup |= CTRST_TOUCHUP;
      computeContrastTable(table, iContrast);
      /*
      | Table[0] must be more or less positive, and table[255]
      | more or less lesser than 255 when the contrast level is negative.
      */
      if (iContrast < 0)
      {
         table[255] = table[254];
         table[0] = table[1];
      }
   }
   if (iBrightness)
   {
      double brightness = ((float)iBrightness+128.0)/128.0;  // [0.0 ... 2.0]
      touchup |= BRITE_TOUCHUP;
      if (brightness <= 1.0)
      {
         m = (Sint32)(sqrt(brightness) * 0x10000);
         k = 0;
         max = 0xFFFFFF;
      }
      else
      {
         double f;
         max = (Sint32)(0xFF / brightness);
         f = brightness - 1.0;
         f = f * f;
         k = (Sint32)(f * 0xFF0000);
         m = (Sint32)((1.0-f) * brightness * 0x10000);
      }
   }
   #ifdef __GNUC__
   else
   {
      m = k = max = 0;  // remove a gcc warning
   }
   #endif
   SDL_LockSurface(surfIn);
   SDL_LockSurface(surfOut);
   if (fmt->palette)          // #if VIDEO_BITSPERPIXEL == 8
   {
      SDL_Color * colors = fmt->palette->colors;
      while (--h >= 0)
      {
         Sint32 i = w;
         Pixel * in = pixelsIn;
         Pixel * out = pixelsOut;
         switch (touchup)
         {
            case NO_TOUCHUP:
               while (--i >= 0)
                  *out++ = *in++;
               break;
            case BRITE_TOUCHUP:
               while (--i >= 0)
               {
                  SDL_Color color = colors[*in++];
                  Sint32 r = color.r;
                  Sint32 g = color.g;
                  Sint32 b = color.b;
                  *out++ = SDL_MapRGB(fmt, (Uint8)((r > max)? 255 : (((m*r) + k) >> 16)),
                                           (Uint8)((g > max)? 255 : (((m*g) + k) >> 16)),
                                           (Uint8)((b > max)? 255 : (((m*b) + k) >> 16)));
               }
               break;
            case CTRST_TOUCHUP:
               while (--i >= 0)
               {
                  SDL_Color color = colors[*in++];
                  *out++ = SDL_MapRGB(fmt, table[color.r], table[color.g], table[color.b]);
               }
               break;
            default: // case CTRSTBRITE_TOUCHUP:
               while (--i >= 0)
               {
                  SDL_Color color = colors[*in++];
                  Sint32 r = table[color.r];
                  Sint32 g = table[color.g];
                  Sint32 b = table[color.b];
                  *out++ = SDL_MapRGB(fmt, (Uint8)((r > max)? 255 : (((m*r) + k) >> 16)),
                                           (Uint8)((g > max)? 255 : (((m*g) + k) >> 16)),
                                           (Uint8)((b > max)? 255 : (((m*b) + k) >> 16)));
               }
               break;
         }
         pixelsIn  = (Pixel *)((char *)pixelsIn  + pitch);
         pixelsOut = (Pixel *)((char *)pixelsOut + pitch);
      }
   }
   else                  // #else // VIDEO_BITSPERPIXEL != 8
   {
      Uint8 rloss  = fmt->Rloss;
      Uint8 gloss  = fmt->Gloss;
      Uint8 bloss  = fmt->Bloss;
      Uint8 rloss2 = rloss + 16;
      Uint8 gloss2 = gloss + 16;
      Uint8 bloss2 = bloss + 16;
      Uint8 rshift = fmt->Rshift;
      Uint8 gshift = fmt->Gshift;
      Uint8 bshift = fmt->Bshift;
      Uint32 rmask = fmt->Rmask;
      Uint32 gmask = fmt->Gmask;
      Uint32 bmask = fmt->Bmask;
      Uint32 amask = fmt->Amask;

      while (--h >= 0)
      {
         Sint32 i = w;
         Pixel * in = pixelsIn;
         Pixel * out = pixelsOut;
         switch (touchup)
         {
            case NO_TOUCHUP:
               while (--i >= 0)
                  *out++ = *in++;
               break;
            case BRITE_TOUCHUP:
               while (--i >= 0)
               {
                  Pixel px = *in++;
                  Sint32 r = ((px & rmask) >> rshift) << rloss;
                  Sint32 g = ((px & gmask) >> gshift) << gloss;
                  Sint32 b = ((px & bmask) >> bshift) << bloss;
                  *out++ = (((((r > max)? 0xFF0000 : ((m*r) + k)) >> rloss2) << rshift) |
                            ((((g > max)? 0xFF0000 : ((m*g) + k)) >> gloss2) << gshift) |
                            ((((b > max)? 0xFF0000 : ((m*b) + k)) >> bloss2) << bshift) |
                            (amask & px));
               }
               break;
            case CTRST_TOUCHUP:
               while (--i >= 0)
               {
                  Pixel px = *in++;
                  *out++ = (((table[((px & rmask) >> rshift) << rloss] >> rloss) << rshift) |
                            ((table[((px & gmask) >> gshift) << gloss] >> gloss) << gshift) |
                            ((table[((px & bmask) >> bshift) << bloss] >> bloss) << bshift) |
                            (amask & px));
               }
               break;
            default: // case CTRSTBRITE_TOUCHUP:
               while (--i >= 0)
               {
                  Pixel px = *in++;
                  Sint32 r = table[((px & rmask) >> rshift) << rloss];
                  Sint32 g = table[((px & gmask) >> gshift) << gloss];
                  Sint32 b = table[((px & bmask) >> bshift) << bloss];
                  *out++ = (((((r > max)? 0xFF0000 : ((m*r) + k)) >> rloss2) << rshift) |
                            ((((g > max)? 0xFF0000 : ((m*g) + k)) >> gloss2) << gshift) |
                            ((((b > max)? 0xFF0000 : ((m*b) + k)) >> bloss2) << bshift) |
                            (amask & px));
               }
               break;
         }
         pixelsIn  = (Pixel *)((char *)pixelsIn  + pitch);
         pixelsOut = (Pixel *)((char *)pixelsOut + pitch);
      }
   }                           // #endif // VIDEO_BITSPERPIXEL
   SDL_UnlockSurface(surfIn);
   SDL_UnlockSurface(surfOut);
   RETURN32(imageOut);
}
/*===========================================================================*/


#if defined(SUPPORT_PUBLIC_SDL) //fdie@ SDL still doesn't support compressed BMP files

#include "SDL_endian.h"

/* Compression encodings for BMP files */
#ifndef BI_RGB
#define BI_RGB    0
#define BI_RLE8      1
#define BI_RLE4      2
#define BI_BITFIELDS 3
#endif

static int readRlePixels(SDL_Surface * surface, SDL_RWops *src, int isRle8);

SDL_Surface *SW_LoadBMP_RW(SDL_RWops *src, int freesrc)
{
   int was_error;
   long fp_offset;
   int bmpPitch;
   int i, pad;
   SDL_Surface *surface;
   Uint32 Rmask;
   Uint32 Gmask;
   Uint32 Bmask;
   SDL_Palette *palette;
   Uint8 *bits;
   int ExpandBMP;

   /* The Win32 BMP file header (14 bytes) */
   char   magic[2];
   Uint32 bfSize;
   Uint16 bfReserved1;
   Uint16 bfReserved2;
   Uint32 bfOffBits;

   /* The Win32 BITMAPINFOHEADER struct (40 bytes) */
   Uint32 biSize;
   Sint32 biWidth;
   Sint32 biHeight;
   Uint16 biPlanes;
   Uint16 biBitCount;
   Uint32 biCompression;
   Uint32 biSizeImage;
   Sint32 biXPelsPerMeter;
   Sint32 biYPelsPerMeter;
   Uint32 biClrUsed;
   Uint32 biClrImportant;

    _LOG_INFO((LOG_CAT_SDL, "SW_LoadBMP_RW"));

   /* Make sure we are passed a valid data source */
   surface = NULL;
   was_error = 0;
   if ( src == NULL ) {
      was_error = 1;
      goto done;
   }

   /* Read in the BMP file header */
   fp_offset = SDL_RWtell(src);
   SDL_ClearError();
   if ( SDL_RWread(src, magic, 1, 2) != 2 ) {
      SDL_Error(SDL_EFREAD);
      was_error = 1;
      goto done;
   }
   if ( strncmp(magic, "BM", 2) != 0 ) {
      SDL_SetError("File is not a Windows BMP file");
      was_error = 1;
      goto done;
   }
   bfSize      = SDL_ReadLE32(src);
   bfReserved1 = SDL_ReadLE16(src);
   bfReserved2 = SDL_ReadLE16(src);
   bfOffBits   = SDL_ReadLE32(src);

   /* Read the Win32 BITMAPINFOHEADER */
   biSize      = SDL_ReadLE32(src);
   if ( biSize == 12 ) {
      biWidth     = (Uint32)SDL_ReadLE16(src);
      biHeight = (Uint32)SDL_ReadLE16(src);
      biPlanes = SDL_ReadLE16(src);
      biBitCount  = SDL_ReadLE16(src);
      biCompression  = BI_RGB;
      biSizeImage = 0;
      biXPelsPerMeter   = 0;
      biYPelsPerMeter   = 0;
      biClrUsed   = 0;
      biClrImportant = 0;
   } else {
      biWidth     = SDL_ReadLE32(src);
      biHeight = SDL_ReadLE32(src);
      biPlanes = SDL_ReadLE16(src);
      biBitCount  = SDL_ReadLE16(src);
      biCompression  = SDL_ReadLE32(src);
      biSizeImage = SDL_ReadLE32(src);
      biXPelsPerMeter   = SDL_ReadLE32(src);
      biYPelsPerMeter   = SDL_ReadLE32(src);
      biClrUsed   = SDL_ReadLE32(src);
      biClrImportant = SDL_ReadLE32(src);
   }

    _LOG_INFO((LOG_CAT_SDL, "size=%dx%dx%d hdrSize=%d", (int)biWidth, (int)biHeight, (int)biPlanes, (int)biSize));
    _LOG_INFO((LOG_CAT_SDL, "BitCount=%d Compression=%d SizeImage=%d", (int)biBitCount, (int)biCompression, (int)biSizeImage));
    _LOG_INFO((LOG_CAT_SDL, "ClrUsed=%d ClrImportant=%d", (int)biClrUsed, (int)biClrImportant));
    
    _LOG_INFO((LOG_CAT_SDL, "%s endian", SDL_BYTEORDER == SDL_BIG_ENDIAN ? "Big":"Little"));

   /* Check for read error */
   if ( strcmp(SDL_GetError(), "") != 0 ) {
      was_error = 1;
      goto done;
   }

   /* Expand 1 and 4 bit bitmaps to 8 bits per pixel */
   switch (biBitCount) {
      case 1:
      case 4:
         ExpandBMP = biBitCount;
         biBitCount = 8;
         break;
      default:
         ExpandBMP = 0;
         break;
   }

   /* RLE4 and RLE8 BMP compression is supported */
   Rmask = Gmask = Bmask = 0;
   switch (biCompression) {
      case BI_RGB:
         /* If there are no masks, use the defaults */
         if ( bfOffBits == (14+biSize) ) {
            /* Default values for the BMP format */
            switch (biBitCount) {
               case 15:
               case 16:
                  Rmask = 0x7C00;
                  Gmask = 0x03E0;
                  Bmask = 0x001F;
                  break;
               case 24:
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
                       Rmask = 0x000000FF;
                       Gmask = 0x0000FF00;
                       Bmask = 0x00FF0000;
                  break;
#endif
               case 32:
                  Rmask = 0x00FF0000;
                  Gmask = 0x0000FF00;
                  Bmask = 0x000000FF;
                  break;
               default:
                  break;
            }
            break;
         }
         /* Fall through -- read the RGB masks */

      default:
         switch (biBitCount) {
            case 15:
            case 16:
            case 32:
               Rmask = SDL_ReadLE32(src);
               Gmask = SDL_ReadLE32(src);
               Bmask = SDL_ReadLE32(src);
               break;
            default:
               break;
         }
         break;
   }

   /* Create a compatible surface, note that the colors are RGB ordered */
   surface = SDL_CreateRGBSurface(SDL_SWSURFACE,
         biWidth, biHeight, biBitCount, Rmask, Gmask, Bmask, 0);
   if ( surface == NULL ) {
      was_error = 1;
      goto done;
   }

   /* Load the palette, if any */
   palette = (surface->format)->palette;
   if ( palette ) {
      /*
      | guich: always use 1<<bpp b/c some bitmaps can bring wrong information
      | for colorsUsed
      */
      /* if ( biClrUsed == 0 ) {  */
      biClrUsed = 1 << biBitCount;
        _LOG_INFO((LOG_CAT_SDL, "%d palette", (int)biClrUsed));
      /* } */
      if ( biSize == 12 ) {
         for ( i = 0; i < (int)biClrUsed; ++i ) {
            SDL_RWread(src, &palette->colors[i].b, 1, 1);
            SDL_RWread(src, &palette->colors[i].g, 1, 1);
            SDL_RWread(src, &palette->colors[i].r, 1, 1);
            palette->colors[i].unused = 0;
                _LOG_DEBUG((LOG_CAT_SDL, "+pal(%d)=0x%x,0x%x,0x%x", (int)i, 
                            (int)palette->colors[i].r, (int)palette->colors[i].g, (int)palette->colors[i].b));
         }  
      } else {
         for ( i = 0; i < (int)biClrUsed; ++i ) {
            SDL_RWread(src, &palette->colors[i].b, 1, 1);
            SDL_RWread(src, &palette->colors[i].g, 1, 1);
            SDL_RWread(src, &palette->colors[i].r, 1, 1);
            SDL_RWread(src, &palette->colors[i].unused, 1, 1);
                _LOG_DEBUG((LOG_CAT_SDL, "*pal(%d)=0x%x,0x%x,0x%x", (int)i, 
                            (int)palette->colors[i].r, (int)palette->colors[i].g, (int)palette->colors[i].b));
         }  
      }
      palette->ncolors = biClrUsed;
   }

   /* Read the surface pixels.  Note that the bmp image is upside down */
   if ( SDL_RWseek(src, fp_offset+bfOffBits, SEEK_SET) < 0 ) {
      SDL_Error(SDL_EFSEEK);
      was_error = 1;
      goto done;
   }
   if ((biCompression == BI_RLE4) || (biCompression == BI_RLE8)) {
      SDL_LockSurface(surface);   /* petrus: not sure if I have to this.  Sam? */
        _LOG_INFO((LOG_CAT_SDL, "readRlePixels"));
      was_error = readRlePixels(surface, src, biCompression == BI_RLE8);
      SDL_UnlockSurface(surface); /* see comment above. */
      if (was_error) SDL_SetError("Error reading from BMP");
      goto done;
   }
   bits = (Uint8 *)surface->pixels+(surface->h*surface->pitch);
    _LOG_DEBUG((LOG_CAT_SDL, "ExpandBMP=%d", (int)ExpandBMP));
   switch (ExpandBMP) {
      case 1:
         bmpPitch = (biWidth + 7) >> 3;
         pad  = (((bmpPitch)%4) ? (4-((bmpPitch)%4)) : 0);
         break;
      case 4:
         bmpPitch = (biWidth + 1) >> 1;
         pad  = (((bmpPitch)%4) ? (4-((bmpPitch)%4)) : 0);
         break;
      default:
         pad  = ((surface->pitch%4) ?
               (4-(surface->pitch%4)) : 0);
         break;
   }
    _LOG_DEBUG((LOG_CAT_SDL, "bmpPitch=%d, pad=%d", (int)bmpPitch, (int)pad));

   while ( bits > (Uint8 *)surface->pixels ) {
      bits -= surface->pitch;
        _LOG_DEBUG((LOG_CAT_SDL, "readline: 0x%x", bits));
        
      switch (ExpandBMP) {
         case 1:
         case 4: {
         Uint8 pixel = 0;
         int   shift = (8-ExpandBMP);
         for ( i=0; i<surface->w; ++i ) {
            if ( i%(8/ExpandBMP) == 0 ) {
               if ( !SDL_RWread(src, &pixel, 1, 1) ) {
                  SDL_SetError("Error reading from BMP");
                  was_error = 1;
                  goto done;
               }
            }
            *(bits+i) = (pixel>>shift);
            pixel <<= ExpandBMP;
         } }
         break;

         default:
         if ( SDL_RWread(src, bits, 1, surface->pitch) != surface->pitch ) {
            SDL_Error(SDL_EFREAD);
            was_error = 1;
            goto done;
         }
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
         /* Byte-swap the pixels if needed. Note that the 24bpp
            case has already been taken care of above. */
         switch(biBitCount) {
            case 15:
            case 16: {
                    Uint16 *pix = (Uint16 *)bits;
               for(i = 0; i < surface->w; i++)
                       pix[i] = SDL_Swap16(pix[i]);
               break;
            }

            case 32: {
                    Uint32 *pix = (Uint32 *)bits;
               for(i = 0; i < surface->w; i++)
                       pix[i] = SDL_Swap32(pix[i]);
               break;
            }
         }
#endif
         break;
      }
      /* Skip padding bytes, ugh */
      if ( pad ) {
         Uint8 padbyte;
         for ( i=0; i<pad; ++i ) {
            SDL_RWread(src, &padbyte, 1, 1);
         }
      }
   }
done:
   if ( was_error ) {
      if ( surface ) {
         SDL_FreeSurface(surface);
      }
      surface = NULL;
   }
   if ( freesrc && src ) {
      SDL_RWclose(src);
   }
   return(surface);
}

static int readRlePixels(SDL_Surface * surface, SDL_RWops * src, int isRle8)
{
   /*
   | Sets the surface pixels from src.  A bmp image is upside down.
   */
   int pitch = surface->pitch;
   int height = surface->h;
   Uint8 * bits = (Uint8 *)surface->pixels + ((height-1) * pitch);
   int ofs = 0;
   Uint8 ch;
   Uint8 needsPad;

   for (;;) {
      if ( !SDL_RWread(src, &ch, 1, 1) ) return 1;
      /*
      | encoded mode starts with a run length, and then a byte
      | with two colour indexes to alternate between for the run
      */
      if ( ch ) {
         Uint8 pixel;
         if ( !SDL_RWread(src, &pixel, 1, 1) ) return 1;
         if ( isRle8 ) {                 /* 256-color bitmap, compressed */
            do {
               bits[ofs++] = pixel;
            } while (--ch);
         }else {                         /* 16-color bitmap, compressed */
            Uint8 pixel0 = pixel >> 4;
            Uint8 pixel1 = pixel & 0x0F;
            for (;;) {
               bits[ofs++] = pixel0;     /* even count, high nibble */
               if (!--ch) break;
               bits[ofs++] = pixel1;     /* odd count, low nibble */
               if (!--ch) break;
            }
         }
      } else {
         /*
         | A leading zero is an escape; it may signal the end of the bitmap,
         | a cursor move, or some absolute data.
         | zero tag may be absolute mode or an escape
         */
         if ( !SDL_RWread(src, &ch, 1, 1) ) return 1;
         switch (ch) {
         case 0:                         /* end of line */
            ofs = 0;
            bits -= pitch;               /* go to previous */
            break;
         case 1:                         /* end of bitmap */
            return 0;                    /* success! */
         case 2:                         /* delta */
            if ( !SDL_RWread(src, &ch, 1, 1) ) return 1;
            ofs += ch;
            if ( !SDL_RWread(src, &ch, 1, 1) ) return 1;
            bits -= (ch * pitch);
            break;
         default:                        /* no compression */
            if (isRle8) {
               needsPad = ( ch & 1 );
               do {
                  if ( !SDL_RWread(src, bits + ofs++, 1, 1) ) return 1;
               } while (--ch);
            } else {
               needsPad = ( ((ch+1)>>1) & 1 ); /* (ch+1)>>1: bytes size */
               for (;;) {
                  Uint8 pixel;
                  if ( !SDL_RWread(src, &pixel, 1, 1) ) return 1;
                  bits[ofs++] = pixel >> 4;
                  if (!--ch) break;
                  bits[ofs++] = pixel & 0x0F;
                  if (!--ch) break;
               }
            }
            /* pad at even boundary */
            if ( needsPad && !SDL_RWread(src, &ch, 1, 1) ) return 1;
            break;
         }
      }
   }
}

#endif
